import http.server
import logging
import textwrap
from itertools import groupby
from operator import itemgetter
from pathlib import Path
from typing import Dict, Union

import numpy as np
from Bio.SeqUtils import seq1
from jinja2 import Environment, FileSystemLoader

from abcfold.output.alphafold3 import AlphafoldOutput
from abcfold.output.boltz import BoltzOutput
from abcfold.output.chai import ChaiOutput
from abcfold.output.file_handlers import ConfidenceJsonFile, NpzFile
from abcfold.output.protenix import ProtenixOutput
from abcfold.plots.pae_plot import create_pae_plots
from abcfold.plots.plddt_plot import plot_plddt
from abcfold.scripts.ipsae import Ipsae

PORT = 8000
logger = logging.getLogger("logger")


def get_plddt_regions(plddts: Union[np.ndarray, list]) -> dict:
    """
    Get the pLDDT regions for the model
    """
    if not isinstance(plddts, np.ndarray):
        plddts = np.array(plddts)

    regions = {}
    # replace none values with -1
    plddts = np.where(plddts == None, -1, plddts) # noqa F401

    v_low = np.where((0 <= plddts) & (plddts <= 50))[0]
    regions["v_low"] = get_regions_helper(v_low)
    low = np.where((plddts > 50) & (plddts < 70))[0]
    regions["low"] = get_regions_helper(low)
    confident = np.where((plddts >= 70) & (plddts < 90))[0]
    regions["confident"] = get_regions_helper(confident)
    v_confident = np.where(plddts >= 90)[0]
    regions["v_high"] = get_regions_helper(v_confident)

    return regions


def get_regions_helper(indices):
    """
    Get the regions from the indices
    """
    regions = []
    for _, g in groupby(enumerate(indices), lambda x: x[0] - x[1]):
        group = map(itemgetter(1), g)
        group = list(map(int, group))
        regions.append((group[0], group[-1]))
    return regions


def get_model_sequence_data(cif_objs) -> dict:
    """
    Get the sequence for each chain and ligand in the model, used internally
    for plotting

    Args:
        cif_objs : A list of CifFile objs

    Returns:
        dict : Chain ID and sequence data
    """
    sequence_data: dict = {}
    for cif_obj in cif_objs:
        sequence_data_ = {}
        for chain in cif_obj.get_chains():
            if cif_obj.check_ligand(chain):
                if chain.id not in sequence_data_:
                    sequence_data_[chain.id] = ""
                sequence_data_[chain.id] += "".join(
                    [atom.id[0] for residue in chain for atom in residue]
                )
            elif cif_obj.check_other(chain, ["dna"]):
                sequence_data_[chain.id] = "".join(
                    [residue.get_resname()[-1] for residue in chain]
                )
            elif cif_obj.check_other(chain, ["rna"]):
                sequence_data_[chain.id] = "".join(
                    [residue.get_resname() for residue in chain]
                )
            else:

                sequence_data_[chain.id] = "".join(
                    [seq1(residue.get_resname()) for residue in chain]
                )
        sequence_data = {
            chain_id: sorted(
                [sequence_data_[chain_id], sequence_data.get(chain_id, "")],
                reverse=True,
            )[0]
            for chain_id in sequence_data_
        }

    return sequence_data


def get_model_data(model,
                   plot_dict,
                   method,
                   plddt_scores,
                   pae_obj,
                   score_file,
                   output_dir):
    """
    Get the model data for the output page

    Args:
        model (CifFile): Model object
        plot_dict (dict): Dictionary of plots
        method (str): Method used to generate the model
        score_file (str): Path to the file containing model scores
        pae_obj (ConfidenceJsonFile): Obj containing PAE data
        output_dir (Path): Path to the output directory
    """
    regions = get_plddt_regions(plddt_scores)
    ptm_score, iptm_score = parse_scores(score_file)
    full_model_path = Path(model.pathway)
    model_path = full_model_path.relative_to(output_dir).as_posix()

    ipsae_score = []
    try:
        ipsae = Ipsae(model, pae_obj)
        distances = ipsae.get_cb_distance()
        ipsae_scores = ipsae.compute_iptm_ipsae(distances)

        # Currently displaying all chain variations, might alter to max in future
        for k, v in ipsae_scores["ipsae_d0res_asym"].items():
            for k2, v2 in v.items():
                ipsae_score.append(f"{k}{k2}:{np.round(v2, 4)}")

        # Output full ipsae scores for each model as we are only outputting d0res_asym
        ipsae_out = full_model_path.parent / f"{Path(model_path).stem}_ipsae.csv"
        ipsae.output_results(pdockq_scores=None,
                             pdockq2_scores=None,
                             lis_scores=None,
                             ipsae_scores=ipsae_scores,
                             verbose=False,
                             output_csv=ipsae_out)
    except ValueError:
        logger.error("ValueError when calculating ipSAE score, bypassing ipSAE")
        pass

    model_data = {
        "model_id": model.name,
        "model_source": method,
        "model_path": model_path,
        "plddt_regions": regions,
        "avg_plddt": model.average_plddt,
        "h_score": model.h_score,
        "ptm_score": ptm_score,
        "iptm_score": iptm_score,
        "ipsae_score": ",".join(ipsae_score),
        "residue_clashes": model.clashes_residues,
        "atom_clashes": model.clashes,
        "pae_path": Path(plot_dict[model.pathway.as_posix()])
        .relative_to(output_dir)
        .as_posix(),
    }
    return model_data


class NoCacheHTTPRequestHandler(http.server.SimpleHTTPRequestHandler):
    def end_headers(self):
        self.send_header(
            "Cache-Control", "no-store, no-cache, must-revalidate, max-age=0"
        )
        self.send_header("Pragma", "no-cache")
        self.send_header("Expires", "0")
        super().end_headers()


def plots(outputs: list, output_dir: Path):
    """
    Generate plots for the output of the different programs

    Args:
        outputs (list): List of output objects

    """
    pathway_plots = create_pae_plots(outputs, output_dir=output_dir)
    plddt_plot_input: Dict[str, list] = get_all_cif_files(outputs)

    plot_plddt(plddt_plot_input, output_name=output_dir.joinpath("plddt_plot.html"))

    pathway_plots["plddt"] = str(output_dir.joinpath("plddt_plot.html").resolve())

    return pathway_plots


def render_template(in_file_path, out_file_path, **kwargs):
    """
    Templates the given file with the keyword arguments.

    Args:
        in_file_path (Path): The path to the template.
        out_file_path (Path): The path to output the templated file.
        **kwargs (dict): Variables to use in templating.
    """
    env = Environment(
        loader=FileSystemLoader(in_file_path.parent), keep_trailing_newline=True
    )
    template = env.get_template(in_file_path.name)
    output = template.render(**kwargs)
    with open(str(out_file_path), "w") as f:
        f.write(output)


def output_open_html_script(file_out: str, port: int = 8000):
    """
    Make a python script to open the output HTML file in the default web browser

    Args:
        file_out (str): Path to the output script
        port (int): Port to run the server on
    """

    script = f"""
    import http.server
    import socketserver
    import webbrowser
    import sys

    PORT = {port}

    class NoCacheHTTPRequestHandler(http.server.SimpleHTTPRequestHandler):
        def end_headers(self):
            self.send_header("Cache-Control",
                            "no-store, no-cache, must-revalidate, max-age=0")
            self.send_header("Pragma", "no-cache")
            self.send_header("Expires", "0")
            super().end_headers()

    try:
        with socketserver.TCPServer(("", PORT),
                                    NoCacheHTTPRequestHandler) as httpd:
            print(
                f"Serving at port {PORT}: http://localhost:{PORT}/index.html"
                )
            print("Press Ctrl+C to stop the server")
            webbrowser.open(f"http://localhost:{PORT}/index.html")
            httpd.serve_forever()
    except KeyboardInterrupt:
        print("Server stopped")
        httpd.server_close()
        sys.exit(0)
    """

    script = textwrap.dedent(script)
    with open(file_out, "w") as f:
        f.write(script)


def get_all_cif_files(outputs) -> Dict[str, list]:
    method_cif_objs: Dict[str, list] = {}

    for output in outputs:
        if isinstance(output, AlphafoldOutput):
            for seed in output.seeds:
                if "Alphafold3" not in method_cif_objs:
                    method_cif_objs["Alphafold3"] = []
                method_cif_objs["Alphafold3"].extend(output.cif_files[seed])
        elif isinstance(output, BoltzOutput):
            for seed in output.seeds:
                if "Boltz" not in method_cif_objs:
                    method_cif_objs["Boltz"] = []
                method_cif_objs["Boltz"].extend(output.cif_files[seed])
        elif isinstance(output, ChaiOutput):
            for seed in output.seeds:
                if "Chai-1" not in method_cif_objs:
                    method_cif_objs["Chai-1"] = []
                method_cif_objs["Chai-1"].extend(output.cif_files[seed])
        elif isinstance(output, ProtenixOutput):
            for seed in output.seeds:
                if "Protenix" not in method_cif_objs:
                    method_cif_objs["Protenix"] = []
                method_cif_objs["Protenix"].extend(output.cif_files[seed])

    return method_cif_objs


def parse_scores(score_file: Union[ConfidenceJsonFile, NpzFile]) -> tuple:
    """
    Parse the scores from the score file.

    Args:
        score_file (Union[ConfidenceJsonFile, NpzFile]): The score file object.

    Returns:
        tuple: A tuple containing ptm_score and iptm_score as floats, or None if invalid
    """
    ptm_score = None
    iptm_score = None

    if isinstance(score_file, ConfidenceJsonFile):
        data = score_file.load_json_file()
    elif isinstance(score_file, NpzFile):
        data = score_file.load_npz_file()
    else:
        return ptm_score, iptm_score
    for key in ("ptm", "iptm"):
        try:
            value = float(data[key])
            if key == "ptm":
                ptm_score = round(value, 2)
            else:
                iptm_score = round(value, 2)
        except (KeyError, TypeError, ValueError):
            continue

    return ptm_score, iptm_score



import shutil
import tempfile
from pathlib import Path

import pytest

from abcfold.html.html_utils import get_plddt_regions
from abcfold.output import file_handlers


def test_npz_file(test_data):
    test_npz = Path(test_data.test_boltz_6BJ9_seed_1_).joinpath(
        "predictions/test_mmseqs/pae_test_mmseqs_model_1.npz"
    )
    npz_file = file_handlers.NpzFile(test_npz)

    assert isinstance(npz_file.data, dict)
    assert "pae" in npz_file.data.keys()


def test_npy_file(test_data):
    test_npy = Path(test_data.test_chai1_6BJ9_seed_1_).joinpath("pae_scores.npy")
    npy_file = file_handlers.NpyFile(test_npy)

    assert npy_file.data.shape == (2, 888, 888)


def test_cif_file(test_data):
    test_cif = Path(test_data.test_alphafold3_6BJ9_).joinpath(
        "seed-1_sample-0/model.cif"
    )
    cif_file = file_handlers.CifFile(test_cif)
    cif_file.check_clashes()

    assert cif_file.clashes == 0
    assert cif_file.average_plddt == pytest.approx(95.8, rel=1e-2)
    assert cif_file.h_score == 89

    plddt_regions_check = {
        "v_low": [(0, 1), (393, 394)],
        "low": [(2, 2), (395, 395)],
        "confident": [
            (3, 4),
            (23, 23),
            (28, 28),
            (44, 44),
            (49, 49),
            (134, 141),
            (176, 176),
            (180, 180),
            (200, 200),
            (206, 206),
            (208, 214),
            (228, 228),
            (234, 234),
            (240, 240),
            (267, 267),
            (272, 272),
            (330, 330),
            (396, 397),
            (416, 416),
            (421, 421),
            (437, 437),
            (442, 442),
            (527, 534),
            (569, 569),
            (573, 573),
            (593, 593),
            (599, 599),
            (601, 607),
            (610, 610),
            (621, 621),
            (627, 627),
            (633, 633),
            (660, 660),
            (665, 665),
            (723, 723),
        ],
        "v_high": [
            (5, 22),
            (24, 27),
            (29, 43),
            (45, 48),
            (50, 133),
            (142, 175),
            (177, 179),
            (181, 199),
            (201, 205),
            (207, 207),
            (215, 227),
            (229, 233),
            (235, 239),
            (241, 266),
            (268, 271),
            (273, 329),
            (331, 392),
            (398, 415),
            (417, 420),
            (422, 436),
            (438, 441),
            (443, 526),
            (535, 568),
            (570, 572),
            (574, 592),
            (594, 598),
            (600, 600),
            (608, 609),
            (611, 620),
            (622, 626),
            (628, 632),
            (634, 659),
            (661, 664),
            (666, 722),
            (724, 787),
        ],
    }
    plddts = cif_file.residue_plddts
    plddt_regions = get_plddt_regions(plddts)

    print(f"PLDDT regions: {plddt_regions}")

    assert plddt_regions_check == plddt_regions
    assert cif_file.name == "model"
    assert cif_file.pathway == test_cif
    assert len(cif_file.plddts) > len(
        [
            plddts
            for plddts in cif_file.get_plddt_per_residue().values()
            for plddts in plddts
        ]
    )


def test_confidence_json_file(test_data):
    confidence_json = Path(test_data.test_alphafold3_6BJ9_).joinpath(
        "seed-1_sample-0/confidences.json"
    )
    confidence_file = file_handlers.ConfidenceJsonFile(confidence_json)

    assert "atom_chain_ids" in confidence_file.data
    assert "atom_plddts" in confidence_file.data
    assert "contact_probs" in confidence_file.data
    assert "pae" in confidence_file.data


def test_superpose_models(test_data):
    """
    Test the superpose_models function to ensure it correctly superposes two CifFiles.
    """

    with tempfile.TemporaryDirectory() as temp_dir:
        model_1 = Path(test_data.test_alphafold3_6BJ9_).joinpath(
            "seed-1_sample-0/model.cif"
        )
        model_2 = Path(test_data.test_boltz_6BJ9_seed_1_).joinpath(
            "predictions/test_mmseqs/test_mmseqs_model_0.cif"
        )

        shutil.copyfile(model_1, f"{temp_dir}/model_1.cif")
        shutil.copyfile(model_2, f"{temp_dir}/model_2.cif")

        models = [Path(f"{temp_dir}/model_1.cif"), Path(f"{temp_dir}/model_2.cif")]
        file_handlers.superpose_models(models)

        superposed_atoms = []
        count = 0
        with open(f"{temp_dir}/model_2.cif", "r") as f:
            for line in f:
                if line.startswith("ATOM"):
                    fields = line.split()
                    if count <= 10:
                        superposed_atoms.append([fields[10], fields[11], fields[12]])
                    count += 1

        ref_superposed_atoms = [
            ["-4.804", "-5.203", "18.445"],
            ["-4.255", "-5.653", "19.697"],
            ["-3.629", "-7.029", "19.656"],
            ["-4.233", "-7.987", "19.137"],
            ["-2.464", "-7.105", "20.243"],
            ["-1.709", "-8.355", "20.246"],
            ["-1.191", "-8.799", "21.616"],
            ["-1.343", "-9.963", "21.992"],
            ["-0.512", "-8.272", "19.291"],
            ["0.309", "-7.153", "19.645"],
            ["-1.015", "-8.051", "17.846"],
        ]

        print(f"Superposed atoms: {superposed_atoms}")

        assert ref_superposed_atoms == superposed_atoms



import os
import tempfile
from pathlib import Path

import pytest

from abcfold.protenix.af3_to_protenix import ProtenixJson
from abcfold.protenix.run_protenix import (generate_protenix_command,
                                           run_protenix)


@pytest.mark.skipif(os.getenv("CI") == "true", reason="Skipping test in CI environment")
def test_run_protenix(test_data):

    with tempfile.TemporaryDirectory() as temp_dir:
        try:
            run_protenix(
                test_data.test_inputA_json,
                temp_dir,
                save_input=True,
                test=True,
            )
        except Exception as e:
            print(e)
            assert False


def test_generate_protenix_command(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:

        # Need protenix style json file to check whether to use MSA flag
        protenix_json = ProtenixJson(temp_dir)
        protenix_json.json_to_json(test_data.test_inputAmsa_json)
        protenix_json_path = Path(temp_dir) / "protenix1.json"
        protenix_json.write_json(protenix_json_path)

        output_dir = "/road/to/nowhere"

        cmd = generate_protenix_command(
            input_json=protenix_json_path,
            output_dir=output_dir,
            number_of_models=5,
            num_recycles=3,
            seed=42
        )

        assert "runner.inference" in cmd
        assert "--input_json_path" in cmd
        assert protenix_json_path.as_posix() in cmd
        assert "--dump_dir" in cmd
        assert output_dir in cmd
        assert "--sample_diffusion.N_sample" in cmd
        assert "5" in cmd
        assert "--model.N_cycle" in cmd
        assert "3" in cmd
        assert "--seeds" in cmd
        assert "42" in cmd
        assert "--use_msa" in cmd
        assert "True" in cmd



import configparser
import json
import os
import re
import shutil
import socketserver
import sys
import tempfile
import webbrowser
from pathlib import Path

from abcfold.alphafold3.run_alphafold3 import run_alphafold3
from abcfold.argparse_utils import (alphafold_argparse_util,
                                    boltz_argparse_util, chai_argparse_util,
                                    custom_template_argpase_util,
                                    main_argpase_util, mmseqs2_argparse_util,
                                    prediction_argparse_util,
                                    protenix_argparse_util,
                                    raise_argument_errors,
                                    visuals_argparse_util)
from abcfold.html.html_utils import (PORT, NoCacheHTTPRequestHandler,
                                     get_all_cif_files, get_model_data,
                                     get_model_sequence_data,
                                     output_open_html_script, plots,
                                     render_template)
from abcfold.output.alphafold3 import AlphafoldOutput
from abcfold.output.boltz import BoltzOutput
from abcfold.output.chai import ChaiOutput
from abcfold.output.file_handlers import superpose_models
from abcfold.output.protenix import ProtenixOutput
from abcfold.output.utils import (get_gap_indicies, insert_none_by_minus_one,
                                  make_dummy_m8_file)
from abcfold.scripts.abc_script_utils import (check_input_json, make_dir,
                                              make_dummy_af3_db, setup_logger)
from abcfold.scripts.add_custom_template import add_custom_template
from abcfold.scripts.add_mmseqs_msa import add_msa_to_json

logger = setup_logger()

HTML_DIR = Path(__file__).parent / "html"
HTML_TEMPLATE = HTML_DIR.joinpath("abcfold.html.jinja2")
PLOTS_DIR = ".plots"


def run(args, config, defaults, config_file):
    """Run ABCFold

    Args:
        args (argparse.Namespace): Arguments from the command line
        config (configparser.SafeConfigParser): Config parser object
        defaults (dict): Default values from the config file
        config_file (Path): Path to the config file


    Raises:
        SystemExit: If the database directory or model parameters directory is not found


    """
    outputs = []

    args.output_dir = Path(args.output_dir)

    if args.mmseqs2:
        logger.info("MMSeqs2 Selected, all other MSAs will be ignored")

    make_dir(args.output_dir, overwrite=args.override)
    make_dir(args.output_dir.joinpath(PLOTS_DIR))

    updated_config = False
    if args.model_params is not None and args.model_params != defaults["model_params"]:
        config.set("Databases", "model_params", args.model_params)
        updated_config = True
    if args.database_dir is not None and args.database_dir != defaults["database_dir"]:
        config.set("Databases", "database_dir", args.database_dir)
        updated_config = True
    af3_sif = args.af3_sif_path
    if af3_sif is not None and af3_sif != defaults.get("af3_sif_path"):
        config.set("Sif_paths", "af3_sif_path", af3_sif)
        updated_config = True
    if updated_config:
        with open(config_file, "w") as f:
            config.write(f)

    args = raise_argument_errors(args)
    # Ensure that the input json file is valid
    args.input_json = check_input_json(
        args.input_json,
        output_dir=args.output_dir,
        use_af3_templates=args.use_af3_template_search,
    )

    with open(args.input_json, "r") as f:
        input_params = json.load(f)

    name = input_params.get("name")
    if name is None:
        logger.error("Input JSON must contain a 'name' field")
        sys.exit(1)

    with tempfile.TemporaryDirectory() as temp_dir_str:
        temp_dir = Path(temp_dir_str)
        if args.mmseqs2:
            if not args.output_json:
                input_json = Path(args.input_json)
                run_json = temp_dir.joinpath(
                    input_json.name.replace(".json", "_mmseqs.json")
                )
            else:
                input_json = Path(args.input_json)
                run_json = Path(args.output_json)

            input_params = add_msa_to_json(
                input_json=input_json,
                mmseqs_db=args.mmseqs_database,
                templates=args.templates,
                num_templates=args.num_templates,
                chai_template_output=temp_dir.joinpath("all_chains.m8"),
                input_params=input_params,
                output_json=run_json,
                to_file=True,
            )
        else:
            run_json = Path(args.input_json)

        if args.custom_template is not None:
            add_custom_template(
                run_json,
                args.target_id,
                args.custom_template,
                args.custom_template_chain,
                output_json=run_json,
                to_file=True
            )

        successful_runs = []
        if args.alphafold3:
            af3_database = args.database_dir
            if args.mmseqs2 or (
                any(
                    seq.get("protein", {}).get("unpairedMsa")
                    or seq.get("protein", {}).get("unpairedMsaPath")
                    for seq in input_params["sequences"]
                )
            ):
                af3_database = make_dummy_af3_db(temp_dir)

            af3_success = run_alphafold3(
                input_json=run_json,
                output_dir=args.output_dir,
                model_params=args.model_params,
                database_dir=af3_database,
                number_of_models=args.number_of_models,
                num_recycles=args.num_recycles,
                sif_path=af3_sif,
                save_distogram=args.save_distogram
            )

            if af3_success:
                pattern = re.compile(f"^{name}.*", re.IGNORECASE)
                af3_out_dir = list(
                    [
                        dir_
                        for dir_ in args.output_dir.glob("*")
                        if dir_.is_dir() and pattern.match(dir_.name)
                    ]
                )[0]
                ao = AlphafoldOutput(af3_out_dir, input_params, name)
                outputs.append(ao)
                run_json = ao.input_json
            successful_runs.append(af3_success)

        if args.boltz:
            from abcfold.boltz.run_boltz import run_boltz

            boltz_success = run_boltz(
                input_json=run_json,
                output_dir=args.output_dir,
                save_input=args.save_input,
                number_of_models=args.number_of_models,
                num_recycles=args.num_recycles,
            )

            if boltz_success:
                bolt_out_dirs = list(args.output_dir.glob("boltz_results*"))
                bo = BoltzOutput(bolt_out_dirs, input_params, name, args.save_input)
                outputs.append(bo)
            successful_runs.append(boltz_success)

        if args.chai1:
            from abcfold.chai1.run_chai1 import run_chai

            template_hits_path = None
            if args.templates and args.mmseqs2:
                template_hits_path = temp_dir.joinpath("all_chain.m8")
            elif args.templates:
                template_hits_path = make_dummy_m8_file(run_json, temp_dir)

            chai_success = run_chai(
                input_json=run_json,
                output_dir=args.output_dir,
                save_input=args.save_input,
                number_of_models=args.number_of_models,
                num_recycles=args.num_recycles,
                template_hits_path=template_hits_path,
            )

            if chai_success:
                chai_output_dirs = list(args.output_dir.glob("chai_output*"))
                co = ChaiOutput(chai_output_dirs, input_params, name, args.save_input)
                outputs.append(co)
            successful_runs.append(chai_success)

        if args.protenix:
            from abcfold.protenix.run_protenix import run_protenix

            protenix_success = run_protenix(
                input_json=run_json,
                output_dir=args.output_dir,
                save_input=args.save_input,
                number_of_models=args.number_of_models,
                num_recycles=args.num_recycles,
            )

            if protenix_success:
                protenix_output_dirs = list(args.output_dir.glob("protenix_results*"))
                po = ProtenixOutput(
                    protenix_output_dirs, input_params, name, args.save_input
                )
                outputs.append(po)
            successful_runs.append(protenix_success)

        if args.no_visuals:
            logger.info("Visuals disabled")
            return

        if not any(successful_runs):
            logger.error("No models were generated")
            return

        plot_dict = plots(outputs, args.output_dir.joinpath(PLOTS_DIR))

        # Compile data to make output page
        programs_run = []
        cif_models = [
            cif_file
            for cif_list in get_all_cif_files(outputs).values()
            for cif_file in cif_list
        ]
        indicies = get_gap_indicies(*cif_models)
        index_counter = 0

        alphafold_models = {"models": []}

        if args.alphafold3:
            if af3_success:
                programs_run.append("AlphaFold3")
                for seed in ao.output.keys():
                    for idx in ao.output[seed].keys():
                        model = ao.output[seed][idx]["cif"]
                        model.check_clashes()
                        score_file = ao.output[seed][idx]["summary"]
                        plddt = model.residue_plddts
                        pae = ao.output[seed][idx]["af3_pae"]
                        if len(indicies) > 0:
                            plddt = insert_none_by_minus_one(
                                indicies[index_counter], plddt
                            )
                        index_counter += 1
                        model_data = get_model_data(
                            model,
                            plot_dict,
                            "AlphaFold3",
                            plddt,
                            pae,
                            score_file,
                            args.output_dir,
                        )
                        alphafold_models["models"].append(model_data)

        boltz_models = {"models": []}
        if args.boltz:
            if boltz_success:
                programs_run.append("Boltz")
                for seed in bo.output.keys():
                    for idx in bo.output[seed].keys():
                        model = bo.output[seed][idx]["cif"]
                        model.check_clashes()
                        score_file = bo.output[seed][idx]["json"]
                        plddt = model.residue_plddts
                        pae = bo.output[seed][idx]["af3_pae"]
                        if len(indicies) > 0:
                            plddt = insert_none_by_minus_one(
                                indicies[index_counter], plddt
                            )
                        index_counter += 1
                        model_data = get_model_data(
                            model,
                            plot_dict,
                            "Boltz",
                            plddt,
                            pae,
                            score_file,
                            args.output_dir
                        )
                        boltz_models["models"].append(model_data)

        chai_models = {"models": []}
        if args.chai1:
            if chai_success:
                programs_run.append("Chai-1")
                for seed in co.output.keys():
                    for idx in co.output[seed].keys():
                        if idx >= 0:
                            model = co.output[seed][idx]["cif"]
                            model.check_clashes()
                            score_file = co.output[seed][idx]["scores"]
                            plddt = model.residue_plddts
                            pae = co.output[seed][idx]["af3_pae"]
                            if len(indicies) > 0:
                                plddt = insert_none_by_minus_one(
                                    indicies[index_counter], plddt
                                )
                            index_counter += 1
                            model_data = get_model_data(
                                model,
                                plot_dict,
                                "Chai-1",
                                plddt,
                                pae,
                                score_file,
                                args.output_dir,
                            )
                            chai_models["models"].append(model_data)

        protenix_models = {"models": []}
        if args.protenix:
            if protenix_success:
                programs_run.append("Protenix")
                for seed in po.output.keys():
                    for idx in po.output[seed].keys():
                        if idx >= 0:
                            model = po.output[seed][idx]["cif"]
                            model.check_clashes()
                            score_file = po.output[seed][idx]["scores"]
                            plddt = model.residue_plddts
                            pae = po.output[seed][idx]["af3_pae"]
                            if len(indicies) > 0:
                                plddt = insert_none_by_minus_one(
                                    indicies[index_counter], plddt
                                )
                            index_counter += 1
                            model_data = get_model_data(
                                model,
                                plot_dict,
                                "Protenix",
                                plddt,
                                pae,
                                score_file,
                                args.output_dir,
                            )
                            protenix_models["models"].append(model_data)

        combined_models = (
            alphafold_models["models"] +
            boltz_models["models"] +
            chai_models["models"] +
            protenix_models["models"]
        )

        # Make the output directory for the models
        os.makedirs(args.output_dir.joinpath("output_models"), exist_ok=True)
        output_models = []
        for model in combined_models:
            cif_file = args.output_dir.joinpath(model["model_path"])
            if model["model_source"] == "AlphaFold3":
                output_name = "af3_model_" + model["model_id"][-1] + ".cif"
            elif model["model_source"] == "Boltz":
                output_name = "boltz_model_" + model["model_id"][-1] + ".cif"
            elif model["model_source"] == "Chai-1":
                output_name = "chai_model_" + model["model_id"][-1] + ".cif"
            elif model["model_source"] == "Protenix":
                output_name = "protenix_model_" + model["model_id"][-1] + ".cif"
            shutil.copy(
                cif_file,
                args.output_dir.joinpath("output_models").joinpath(output_name),
            )
            output_models.append(
                args.output_dir.joinpath("output_models").joinpath(output_name)
            )
        # Superpose the models
        if len(output_models) > 1:
            superpose_models(output_models)

        sequence_data = get_model_sequence_data(cif_models)
        sequence = ""
        for key in sequence_data.keys():
            sequence += sequence_data[key]
        chain_data = {}
        ref = 0
        for key in sequence_data.keys():
            chain_data["Chain " + key] = (ref, len(sequence_data[key]) + ref - 1)
            ref += len(sequence_data[key])
        results_dict = {
            "sequence": sequence,
            "models": combined_models,
            "plotly_path": Path(plot_dict["plddt"])
            .relative_to(args.output_dir.resolve())
            .as_posix(),
            "chain_data": chain_data,
        }
        results_json = json.dumps(results_dict)

        if not args.output_dir.joinpath(".feature_viewer").exists():
            shutil.copytree(HTML_DIR, args.output_dir / ".feature_viewer")

        if len(programs_run) > 1:
            programs = (
                "Structure predictions for: "
                + ", ".join(programs_run[:-1])
                + " and "
                + programs_run[-1]
            )
        else:
            programs = "Structure predictions for: " + programs_run[0]

        # Create the index page
        HTML_OUT = args.output_dir.joinpath("index.html")
        html_out = Path(HTML_OUT).resolve()
        render_template(
            HTML_TEMPLATE,
            html_out,
            # kwargs appear as variables in the template
            abcfold_html_dir=".feature_viewer",
            programs=programs,
            results_json=results_json,
            version=0.1,
        )
        logger.info(f"Output page written to {HTML_OUT}")

        # Change to the output directory to run the server
        os.chdir(args.output_dir)

        # Make a script to open the output HTML file in the default web browser
        output_open_html_script("open_output.py", port=PORT)

        if args.no_server:
            logger.info("Server disabled")
            logger.info(
                "Run 'python open_output.py' in the output directory to \
view the output pages"
            )
            return

        try:
            # Start the server
            with socketserver.TCPServer(("", PORT), NoCacheHTTPRequestHandler) as httpd:
                logger.info(
                    f"Serving at port {PORT}: http://localhost:{PORT}/index.html"
                )
                logger.info("Press Ctrl+C to stop the server")
                # Open the main HTML page in the default web browser
                webbrowser.open(f"http://localhost:{PORT}/index.html")
                # Keep the server running
                httpd.serve_forever()
        except KeyboardInterrupt:
            logger.info("Server stopped")
            httpd.server_close()
            sys.exit(0)


def main():
    """
    Run AlphaFold3 / Boltz / Chai-1 / Protenix
    """
    import argparse

    parser = argparse.ArgumentParser(
        description="Run AlphaFold3 / Boltz / Chai-1 / Protenix"
    )

    defaults = {}
    config_file = Path(__file__).parent.joinpath("data", "config.ini")
    config = configparser.SafeConfigParser()

    if config_file.exists():
        config.read(str(config_file))
        defaults.update(dict(config.items("Databases")))
        defaults.update(dict(config.items("Sif_paths")))

    parser = main_argpase_util(parser)
    parser = alphafold_argparse_util(parser)
    parser = boltz_argparse_util(parser)
    parser = chai_argparse_util(parser)
    parser = protenix_argparse_util(parser)
    parser = mmseqs2_argparse_util(parser)
    parser = custom_template_argpase_util(parser)
    parser = prediction_argparse_util(parser)
    parser = visuals_argparse_util(parser)

    parser.set_defaults(**defaults)
    args = parser.parse_args()

    run(
        args,
        config,
        defaults,
        config_file,
    )


if __name__ == "__main__":
    main()



import tempfile
from pathlib import Path

from abcfold.html.html_utils import get_model_sequence_data
from abcfold.plots.pae_plot import create_pae_plots
from abcfold.plots.plddt_plot import plot_plddt


def test_plddt_plot(output_objs):

    af3_files = output_objs.af3_output.cif_files["seed-1"]
    boltz_files = output_objs.boltz_output.cif_files["seed-1"]
    chai_files = output_objs.chai_output.cif_files["seed-1"]
    protenix_files = output_objs.protenix_output.cif_files["seed-1"]

    assert len(af3_files) == len(boltz_files) == len(chai_files) == len(protenix_files)
    plot_files = {
        "Alphafold3": af3_files,
        "Boltz": boltz_files,
        "Chai-1": chai_files,
        "Protenix": protenix_files
    }

    with tempfile.TemporaryDirectory() as temp_dir:
        plot_plddt(
            plot_files,
            output_name=f"{temp_dir}/test.html",
        )

        assert Path(f"{temp_dir}/test.html").exists()


def test_pae_plots(output_objs):
    outputs = [
        output_objs.af3_output,
        output_objs.boltz_output,
        output_objs.chai_output,
        output_objs.protenix_output,
    ]

    with tempfile.TemporaryDirectory() as temp_dir_str:
        temp_dir = Path(temp_dir_str)
        plot_pathways = create_pae_plots(outputs, output_dir=temp_dir)

        assert len(plot_pathways) == 8
        values = [Path(value).name for value in plot_pathways.values()]
        print(values)

        assert "confidences_seed-1_sample-0_af3_pae_plot.html" in values
        assert "confidences_seed-1_sample-1_af3_pae_plot.html" in values
        assert "pae_test_mmseqs_model_0_test_mmseqs_af3_pae_plot.html" in values
        assert "pae_test_mmseqs_model_1_test_mmseqs_af3_pae_plot.html" in values
        assert "pae_scores_model_0_chai1_6BJ9_seed-1_af3_pae_plot.html" in values
        assert "pae_scores_model_1_chai1_6BJ9_seed-1_af3_pae_plot.html" in values
        assert "6BJ9_full_data_sample_0_predictions_af3_pae_plot.html" in values
        assert "6BJ9_full_data_sample_1_predictions_af3_pae_plot.html" in values

        assert (
            any(
                "alphafold3_6BJ9/seed-1_sample-0/model.cif" in x for x in plot_pathways
            )
        )
        assert (
            any(
                "alphafold3_6BJ9/seed-1_sample-1/model.cif" in x for x in plot_pathways
            )
        )
        assert (
            any(
                "boltz_6BJ9_seed-1/predictions/test_mmseqs/test_mmseqs_model_0.cif"
                in x for x in plot_pathways
            )
        )
        assert (
            any(
                "boltz_6BJ9_seed-1/predictions/test_mmseqs/test_mmseqs_model_1.cif"
                in x for x in plot_pathways
            )
        )
        assert (
            any("chai1_6BJ9_seed-1/pred.model_idx_0.cif" in x for x in plot_pathways)
        )
        assert (
            any("chai1_6BJ9_seed-1/pred.model_idx_1.cif" in x for x in plot_pathways)
        )
        assert (
            any("protenix_6BJ9_seed-1/6BJ9/seed_1/predictions/6BJ9_sample_0.cif"
                in x for x in plot_pathways)
        )
        assert (
            any("protenix_6BJ9_seed-1/6BJ9/seed_1/predictions/6BJ9_sample_1.cif"
                in x for x in plot_pathways)
        )

        assert len(list(temp_dir.glob("*.html"))) == 8


def test_get_sequence_data(output_objs):
    af3_files = output_objs.af3_output.cif_files["seed-1"]
    boltz_files = output_objs.boltz_output.cif_files["seed-1"]
    chai_files = output_objs.chai_output.cif_files["seed-1"]
    protenix_files = output_objs.protenix_output.cif_files["seed-1"]

    cif_files = []

    [cif_files.extend(files) for files in [af3_files,
                                           boltz_files,
                                           chai_files,
                                           protenix_files]]

    outputdic = get_model_sequence_data(cif_files)

    assert outputdic == {
        "A": "GTGSRPITDVVFVGAARTPIGSFRSAFNNVPVTVLGREALKGALKNANVKPSLVQEAFIGVVVPSNAGQGPA\
RQVVLGAGCDVSTVVTAVNKMCASGMKAIACAASILQLDLQEMVVAGGMESMSCVPFYLPRGEIPFGGTKLIDGIPRDGLNDVYND\
ILMGACADKVAKQFAITREEQDKYAILSYKRSAAAWKEGIFAKEIIPLEVTQGKKTITVEEDEEYKKVNFEKIPKLKPAFTSEGSV\
TAANASTLNDGAAMVVMTTVDGAKKHGLKPLARMLAYGDAATHPIDFGIAPASVIPKVLKLAGLQIKDIDLWEINEAFAVVPLYTM\
KTLGLDESKVNIHGGAVSLGHPIGMSGARIVGHLVHTLKPGQKGCAAICNGGGGAGGMIIEKL",
        "B": "GTGSRPITDVVFVGAARTPIGSFRSAFNNVPVTVLGREALKGALKNANVKPSLVQEAFIGVVVPSNAGQGPA\
RQVVLGAGCDVSTVVTAVNKMCASGMKAIACAASILQLDLQEMVVAGGMESMSCVPFYLPRGEIPFGGTKLIDGIPRDGLNDVYND\
ILMGACADKVAKQFAITREEQDKYAILSYKRSAAAWKEGIFAKEIIPLEVTQGKKTITVEEDEEYKKVNFEKIPKLKPAFTSEGSV\
TAANASTLNDGAAMVVMTTVDGAKKHGLKPLARMLAYGDAATHPIDFGIAPASVIPKVLKLAGLQIKDIDLWEINEAFAVVPLYTM\
KTLGLDESKVNIHGGAVSLGHPIGMSGARIVGHLVHTLKPGQKGCAAICNGGGGAGGMIIEKL",
        "C": "NCNCCCNNCNCCOCOPOOOCOCOPOOOPOOOCCCCCOCONCCCONCCSCOC",
        "D": "NCNCCCNNCNCCOCOPOOOCOCOPOOOPOOOCCCCCOCONCCCONCCSCOC",
    }



import logging
import subprocess
import tempfile
from pathlib import Path
from typing import Union

from abcfold.boltz.af3_to_boltz import BoltzYaml
from abcfold.boltz.check_install import ensure_boltz_env

logger = logging.getLogger("logger")


def run_boltz(
    input_json: Union[str, Path],
    output_dir: Union[str, Path],
    save_input: bool = False,
    test: bool = False,
    number_of_models: int = 5,
    num_recycles: int = 10,
) -> bool:
    """
    Run Boltz using the input JSON file

    Args:
        input_json (Union[str, Path]): Path to the input JSON file
        output_dir (Union[str, Path]): Path to the output directory
        save_input (bool): If True, save the input yaml file and MSA to the output
        directory
        test (bool): If True, run the test command
        number_of_models (int): Number of models to generate

    Returns:
        Bool: True if the Boltz run was successful, False otherwise

    Raises:
        subprocess.CalledProcessError: If the Boltz command returns an error


    """
    input_json = Path(input_json)
    output_dir = Path(output_dir)

    logger.debug("Checking if boltz is installed")
    env = ensure_boltz_env()

    with tempfile.TemporaryDirectory() as temp_dir:
        working_dir = Path(temp_dir)
        if save_input:
            logger.info("Saving input yaml file and msa to the output directory")
            working_dir = output_dir

        boltz_yaml = BoltzYaml(working_dir)
        boltz_yaml.json_to_yaml(input_json)

        for seed in boltz_yaml.seeds:
            out_file = working_dir.joinpath(f"{input_json.stem}_seed-{seed}.yaml")

            boltz_yaml.write_yaml(out_file)
            logger.info("Running Boltz using seed: %s", seed)
            cmd = (
                generate_boltz_command(
                    out_file,
                    output_dir,
                    number_of_models,
                    num_recycles,
                    seed=seed,
                )
                if not test
                else generate_boltz_test_command()
            )

            try:
                stdout = env.run(cmd, capture_output=True)

                # Check for out-of-memory warnings
                if "WARNING: ran out of memory" in stdout:
                    logger.error("Boltz ran out of memory")
                    return False

            except subprocess.CalledProcessError as e:
                stderr = e.stderr or ""
                if stderr:
                    logger.error(stderr)
                    output_err_file = output_dir / "boltz_error.log"
                    output_err_file.write_text(stderr)
                    logger.error(
                        "Boltz run failed. Error log is in %s", output_err_file
                    )
                else:
                    logger.error("Boltz run failed")
                return False

    logger.info("Boltz run complete")
    logger.info("Output files are in %s", output_dir)
    return True


def generate_boltz_command(
    input_yaml: Union[str, Path],
    output_dir: Union[str, Path],
    number_of_models: int = 5,
    num_recycles: int = 10,
    seed: int = 42,
) -> list:
    """
    Generate the Boltz command

    Args:
        input_yaml (Union[str, Path]): Path to the input YAML file
        output_dir (Union[str, Path]): Path to the output directory
        number_of_models (int): Number of models to generate
        seed (int): Seed for the random number generator

    Returns:
        list: The Boltz command
    """
    return [
        "boltz",
        "predict",
        str(input_yaml),
        "--out_dir",
        str(output_dir),
        "--override",
        "--write_full_pae",
        "--write_full_pde",
        "--diffusion_samples",
        str(number_of_models),
        "--recycling_steps",
        str(num_recycles),
        "--seed",
        str(seed),
    ]


def generate_boltz_test_command() -> list:
    """
    Generate the test command for Boltz

    Args:
        None

    Returns:
        list: The Boltz test command
    """

    return [
        "boltz",
        "predict",
        "--help",
    ]



import json
import logging
import subprocess
import tempfile
from pathlib import Path
from typing import Union

from abcfold.protenix.af3_to_protenix import ProtenixJson
from abcfold.protenix.check_install import ensure_protenix_env

logger = logging.getLogger("logger")


def run_protenix(
    input_json: Union[str, Path],
    output_dir: Union[str, Path],
    save_input: bool = False,
    test: bool = False,
    number_of_models: int = 5,
    num_recycles: int = 10,
) -> bool:
    """
    Run Protenix using the input JSON file

    Args:
        input_json (Union[str, Path]): Path to the input JSON file
        output_dir (Union[str, Path]): Path to the output directory
        save_input (bool): If True, save the input yaml file and MSA to the output
        directory
        test (bool): If True, run the test command
        number_of_models (int): Number of models to generate

    Returns:
        Bool: True if the Protenix run was successful, False otherwise

    Raises:
        subprocess.CalledProcessError: If the Protenix command returns an error


    """
    input_json = Path(input_json)
    output_dir = Path(output_dir)

    logger.debug("Checking if protenix is installed")
    env = ensure_protenix_env()

    with tempfile.TemporaryDirectory() as temp_dir:
        working_dir = Path(temp_dir)
        if save_input:
            logger.info("Saving input json file and msa to the output directory")
            working_dir = output_dir

        protenix_json = ProtenixJson(working_dir)
        protenix_json.json_to_json(input_json)

        for seed in protenix_json.seeds:
            out_file = working_dir.joinpath(f"{input_json.stem}_seed-{seed}.json")

            protenix_json.write_json(out_file)
            protenix_out_dir = output_dir / f"protenix_results_seed-{seed}"
            logger.info("Running Protenix using seed: %s", seed)
            cmd = (
                generate_protenix_command(
                    out_file,
                    protenix_out_dir,
                    number_of_models,
                    num_recycles,
                    seed=seed,
                )
                if not test
                else generate_protenix_test_command()
            )

            try:
                env.run(cmd)
            except subprocess.CalledProcessError as e:
                stderr = e.stderr or ""
                if stderr:
                    if working_dir.exists():
                        output_err_file = working_dir / "protenix_error.log"
                    else:
                        output_err_file = working_dir.parent / "protenix_error.log"
                    output_err_file.write_text(stderr)
                    logger.error(
                        "Protenix run failed. Error log is in %s", output_err_file
                    )
                else:
                    logger.error("Protenix run failed")
                return False

    logger.info("Protenix run complete")
    logger.info("Output files are in %s", output_dir)
    return True


def generate_protenix_command(
    input_json: Union[str, Path],
    output_dir: Union[str, Path],
    number_of_models: int,
    num_recycles: int,
    seed: int,
) -> list:
    """
    Generate the Protenix command

    Args:
        input_json (Union[str, Path]): Path to the input JSON file
        output_dir (Union[str, Path]): Path to the output directory
        number_of_models (int): Number of models to generate
        num_recycles (int): Number of recycles
        seed (int): Random seed

    Returns:
        list: The Protenix command
    """

    # Determine if MSA is present in the input JSON
    use_msa = False
    with open(str(input_json), "r") as f:
        data = json.load(f)
    for key, value in data[0].items():
        if key == "sequences":
            for entry in value:
                if "proteinChain" in entry:
                    if "msa" in entry["proteinChain"]:
                        use_msa = True
                        break

    return [
        "python",
        "-m",
        "runner.inference",
        "--input_json_path",
        str(input_json),
        "--dump_dir",
        str(output_dir),
        "--sample_diffusion.N_sample",
        str(number_of_models),
        "--model.N_cycle",
        str(num_recycles),
        "--seeds",
        str(seed),
        "--use_msa",
        str(use_msa),
        "--need_atom_confidence",
        "True"
    ]


def generate_protenix_test_command() -> list:
    """
    Generate the test command for Protenix

    Args:
        None

    Returns:
        list: The Protenix test command
    """

    return [
        "protenix",
        "predict",
        "--help",
    ]



import os
import tempfile
from pathlib import Path

import pytest

from abcfold.alphafold3.run_alphafold3 import generate_af3_cmd, run_alphafold3
from abcfold.scripts.abc_script_utils import make_dummy_af3_db


def test_generate_af3_command(test_data):
    input_json = Path(test_data.test_inputA_json)
    output_dir = Path("/road/to/nowhere")
    model_params = Path("/road/to/nowhere")
    database_dir = Path("/road/to/nowhere")

    cmd = generate_af3_cmd(
        input_json=input_json,
        output_dir=output_dir,
        model_params=model_params,
        database_dir=database_dir,
        sif_path=None,
        interactive=True,
    )

    assert "docker run -it" in cmd
    assert f"--volume {input_json.parent.resolve()}:/root/af_input" in cmd
    assert f"--volume {output_dir.resolve()}:/root/af_output" in cmd
    assert f"--volume {model_params}:/root/models" in cmd
    assert f"--volume {database_dir}:/root/public_databases" in cmd
    assert "--gpus all" in cmd
    assert "alphafold3" in cmd
    assert f"--json_path=/root/af_input/{input_json.name}" in cmd
    assert "--model_dir=/root/models" in cmd
    assert "--output_dir=/root/af_output" in cmd


def test_generate_af3_singularity_command(test_data):
    input_json = Path(test_data.test_inputA_json)
    output_dir = Path("/road/to/nowhere")
    model_params = Path("/road/to/nowhere")
    database_dir = Path("/road/to/nowhere")
    sif_path = Path("/road/to/nowhere.sif")

    cmd = generate_af3_cmd(
        input_json=input_json,
        output_dir=output_dir,
        model_params=model_params,
        database_dir=database_dir,
        sif_path=sif_path,
        interactive=True,
    )

    assert "singularity exec" in cmd
    assert f"--bind {input_json.parent.resolve()}:/root/af_input" in cmd
    assert f"--bind {output_dir.resolve()}:/root/af_output" in cmd
    assert f"--bind {model_params}:/root/models" in cmd
    assert f"--bind {database_dir}:/root/public_databases" in cmd
    assert f"{sif_path}" in cmd
    assert "python /app/alphafold/run_alphafold.py" in cmd
    assert f"--json_path=/root/af_input/{input_json.name}" in cmd
    assert "--model_dir=/root/models" in cmd
    assert "--output_dir=/root/af_output" in cmd
    assert "--num_diffusion_samples" in cmd
    assert "--num_recycles" in cmd


@pytest.mark.skipif(os.getenv("CI") == "true", reason="Skipping test in CI environment")
def test_run_af3(test_data):
    input_json = Path(test_data.test_inputA_json)
    with tempfile.TemporaryDirectory() as temp_dir_str:
        temp_dir = Path(temp_dir_str)
        output_dir = temp_dir / "af3_output"
        model_params = temp_dir / "af3_output"
        database_dir = make_dummy_af3_db(temp_dir)

        run_alphafold3(
            input_json,
            output_dir,
            model_params,
            database_dir,
            sif_path=None,
            interactive=False,
        )
        assert output_dir.exists()
        assert (output_dir / "af3_error.log").exists()
        with open(output_dir / "af3_error.log", "r") as f:
            error_log = f.read()
            assert "No models matched in /root/models" in error_log



import logging

from abcfold.backend_envs import MicromambaEnv

logger = logging.getLogger("logger")

PROTENIX_VERSION = "0.7.3"
PROTENIX_ENV = "abcfold-protenix-py311"


def ensure_protenix_env():
    env = MicromambaEnv(PROTENIX_ENV)

    # 1. Ensure env exists
    env.create(python_version="3.11")

    # 2. Check installed chai version
    installed = env.get_installed_version("protenix")

    if installed != PROTENIX_VERSION:
        if installed is None:
            logger.info("Protenix not found. Installing version: %s", PROTENIX_ENV)
        else:
            logger.info(
                "Protenix version mismatch (found %s). Installing correct version: %s",
                installed,
                PROTENIX_ENV,
            )
        env.pip_install([f"protenix=={PROTENIX_VERSION}"])
    else:
        logger.info("Protenix is already up-to-date (%s)", PROTENIX_ENV)

    # 3. Ensure runtime deps you *actually* need
    env.ensure_package("numpy")
    env.ensure_package("typer")
    env.ensure_package("matplotlib")

    return env



#!/usr/bin/env python

import gzip
import json
import logging
import math
import os
import random
import shutil
import subprocess
import tarfile
import tempfile
import time
from io import StringIO
from pathlib import Path
from typing import List, Sequence, Union

import pandas as pd
import requests  # type: ignore
from tqdm.autonotebook import tqdm

from abcfold.argparse_utils import (custom_template_argpase_util,
                                    mmseqs2_argparse_util)
from abcfold.scripts.abc_script_utils import (align_and_map,
                                              extract_sequence_from_mmcif,
                                              get_custom_template, get_mmcif)

logger = logging.getLogger("logger")

TQDM_BAR_FORMAT = (
    "{l_bar}{bar}| {n_fmt}/{total_fmt} [elapsed: {elapsed} remaining: {remaining}]"
)

MODULE_OUTPUT_POS = {
    "align":        4,
    "convertalis":  4,
    "expandaln":    5,
    "filterresult": 4,
    "lndb":         2,
    "mergedbs":     2,
    "mvdb":         2,
    "pairaln":      4,
    "result2msa":   4,
    "search":       3,
}


class MMseqs2Exception(Exception):
    def __init__(self):

        msg = "MMseqs2 API is giving errors. Please confirm your input is a valid \
protein sequence. If error persists, please try again an hour later."
        logger.error(msg)
        super().__init__()


def add_msa_to_json(
    input_json,
    mmseqs_db,
    templates,
    num_templates,
    chai_template_output,
    custom_template=None,
    custom_template_chain=None,
    target_id=None,
    input_params=None,
    output_json=None,
    to_file=True,
):

    if input_params is None:
        with open(input_json, "r") as f:
            input_params = json.load(f)
    updated_sequences = []
    for sequence in input_params["sequences"]:
        if "protein" in sequence:
            input_id = sequence["protein"]["id"]
            input_sequence = sequence["protein"]["sequence"]
            with tempfile.TemporaryDirectory() as tmpdir:
                if mmseqs_db:
                    logger.info(f"Running Local MMseqs2 on sequence: {input_sequence}")
                    if templates:
                        a3m_lines, template_files = run_local_mmseqs(
                            input_sequence,
                            Path(tmpdir),
                            use_templates=True,
                            num_templates=num_templates,
                            mmseqs_db=Path(mmseqs_db),
                        )
                    else:
                        a3m_lines = run_local_mmseqs(
                            input_sequence,
                            Path(tmpdir),
                            use_templates=False,
                            mmseqs_db=Path(mmseqs_db),
                        )
                else:
                    logger.info(f"Running MMseqs2 on sequence: {input_sequence}")
                    # Run MMseqs2 to get unpaired MSA
                    if templates:
                        a3m_lines, template_files = run_mmseqs(
                            input_sequence,
                            tmpdir,
                            use_templates=True,
                            num_templates=num_templates,
                        )

                        for i in input_id:
                            table = pd.read_csv(
                                f"{tmpdir}/pdb70.m8",
                                delimiter="\t",
                                header=None,
                                names=[
                                    "query_id",
                                    "subject_id",
                                    "pident",
                                    "length",
                                    "mismatch",
                                    "gapopen",
                                    "query_start",
                                    "query_end",
                                    "subject_start",
                                    "subject_end",
                                    "evalue",
                                    "bitscore",
                                    "comment",
                                ],
                            )

                            table["query_id"] = i

                            if chai_template_output:
                                if os.path.exists(chai_template_output):
                                    table.to_csv(
                                        chai_template_output,
                                        sep="\t",
                                        index=False,
                                        header=False,
                                        mode="a",
                                    )
                                else:
                                    table.to_csv(
                                        chai_template_output,
                                        sep="\t",
                                        index=False,
                                        header=False,
                                    )

                    else:
                        a3m_lines = run_mmseqs(
                            input_sequence,
                            tmpdir,
                            use_templates=False)
                        template_files = []

                sequence["protein"]["templates"] = template_files

                if custom_template:
                    for template in custom_template:
                        if not os.path.exists(template):
                            msg = f"Custom template file {template} not found"
                            logger.critical(msg)
                            raise FileNotFoundError()
                        # Can only add templates to protein sequences, so check if there
                        # are multiple protein sequences in the input json
                        if (
                            len(
                                [
                                    x
                                    for x in input_params["sequences"]
                                    if "protein" in x.keys()
                                ]
                            )
                            > 1
                            and not target_id
                        ):
                            msg = "Multiple sequences found in input json. \
Please specify target id so that custom template can be added to the correct sequence"
                            raise ValueError(msg)

                    if target_id and len(target_id) > 1:
                        if (len(custom_template) != len(target_id)) or (
                            len(custom_template_chain) != len(target_id)
                        ):
                            msg = "If providing templates for multiple targets, the \
number of target ids must match the number of custom templates and custom template \
chains"
                            raise ValueError(msg)
                        custom_templates = zip(
                            target_id, custom_template, custom_template_chain
                        )
                    else:
                        if len(custom_template) != len(custom_template_chain):
                            msg = "Number of custom templates must match the number of \
custom template chains"
                            raise ValueError(msg)
                        # if a single target id is provided, assume all custom templates
                        # are for the same target
                        if target_id:
                            target_ids = [target_id[0]] * len(custom_template)
                        else:
                            target_ids = [None] * len(custom_template)
                        custom_templates = zip(
                            target_ids, custom_template, custom_template_chain
                        )

                    for i in custom_templates:
                        tid, c_tem, c_tem_chn = i
                        sequence = get_custom_template(
                            sequence,
                            tid,
                            c_tem,
                            c_tem_chn,
                        )

                # Add unpaired MSA to the json
                sequence["protein"]["unpairedMsa"] = a3m_lines[0]
                sequence["protein"]["pairedMsa"] = ""
                sequence["protein"]["templates"] = template_files
            updated_sequences.append(sequence)
        else:
            # Capture non-protein sequences
            updated_sequences.append(sequence)
    input_params["sequences"] = updated_sequences

    if to_file:
        if output_json:
            with open(output_json, "w") as f:
                json.dump(input_params, f)
        else:
            output_json = input_json.replace(".json", "_mmseqs.json")
            with open(output_json, "w") as f:
                json.dump(input_params, f)

    return input_params


# Lightly modified code from https://github.com/sokrypton/ColabFold
def run_mmseqs(
    x,
    prefix,
    use_env=True,
    use_filter=True,
    use_templates=False,
    filter=None,
    use_pairing=False,
    host_url="https://a3m.mmseqs.com",
    num_templates=20
) -> Sequence[object]:
    submission_endpoint = "ticket/pair" if use_pairing else "ticket/msa"

    def submit(seqs, mode, N=101):
        n, query = N, ""
        for seq in seqs:
            query += f">{n}\n{seq}\n"
            n += 1

        res = requests.post(
            f"{host_url}/{submission_endpoint}", data={"q": query, "mode": mode}
        )
        try:
            out = res.json()
        except ValueError:
            logger.error(f"Server didn't reply with json: {res.text}")
            out = {"status": "ERROR"}
        return out

    def status(ID):
        res = requests.get(f"{host_url}/ticket/{ID}")
        try:
            out = res.json()
        except ValueError:
            logger.error(f"Server didn't reply with json: {res.text}")
            out = {"status": "ERROR"}
        return out

    def download(ID, path):
        res = requests.get(f"{host_url}/result/download/{ID}")
        with open(path, "wb") as out:
            out.write(res.content)

    # process input x
    seqs = [x] if isinstance(x, str) else x

    # compatibility to old option
    if filter is not None:
        use_filter = filter

    # setup mode
    if use_filter:
        mode = "env" if use_env else "all"
    else:
        mode = "env-nofilter" if use_env else "nofilter"

    if use_pairing:
        mode = ""
        use_templates = False
        use_env = False

    # define path
    path = prefix
    if not os.path.isdir(path):
        os.mkdir(path)

    # call mmseqs2 api
    tar_gz_file = f"{path}/out.tar.gz"
    N, REDO = 101, True

    # deduplicate and keep track of order
    seqs_unique = list(set(seqs))
    Ms = [N + seqs_unique.index(seq) for seq in seqs]
    # lets do it!
    if not os.path.isfile(tar_gz_file):
        TIME_ESTIMATE = 150 * len(seqs_unique)
        with tqdm(total=TIME_ESTIMATE, bar_format=TQDM_BAR_FORMAT) as pbar:
            while REDO:
                pbar.set_description("SUBMIT")

                # Resubmit job until it goes through
                out = submit(seqs_unique, mode, N)
                while out["status"] in ["UNKNOWN", "RATELIMIT"]:
                    sleep_time = 5 + random.randint(0, 5)
                    logger.info(f"Sleeping for {sleep_time}s. Reason: {out['status']}")
                    # resubmit
                    time.sleep(sleep_time)
                    out = submit(seqs_unique, mode, N)

                if out["status"] == "ERROR":
                    raise MMseqs2Exception()

                if out["status"] == "MAINTENANCE":
                    raise MMseqs2Exception()

                # wait for job to finish
                ID, TIME = out["id"], 0
                pbar.set_description(out["status"])
                while out["status"] in ["UNKNOWN", "RUNNING", "PENDING"]:
                    t = 5 + random.randint(0, 5)
                    logger.info(f"Sleeping for {t}s. Reason: {out['status']}")
                    time.sleep(t)
                    out = status(ID)
                    pbar.set_description(out["status"])
                    if out["status"] == "RUNNING":
                        TIME += t
                        pbar.update(n=t)

                if out["status"] == "COMPLETE":
                    if TIME < TIME_ESTIMATE:
                        pbar.update(n=(TIME_ESTIMATE - TIME))
                    REDO = False

                if out["status"] == "ERROR":
                    REDO = False
                    raise MMseqs2Exception()

            # Download results
            download(ID, tar_gz_file)

    # prep list of a3m files
    if use_pairing:
        a3m_files = [f"{path}/pair.a3m"]
    else:
        a3m_files = [f"{path}/uniref.a3m"]
        if use_env:
            a3m_files.append(f"{path}/bfd.mgnify30.metaeuk30.smag30.a3m")

    # extract a3m files
    if any(not os.path.isfile(a3m_file) for a3m_file in a3m_files):
        with tarfile.open(tar_gz_file) as tar_gz:
            tar_gz.extractall(path)

    # gather a3m lines
    a3m_lines: dict = {}
    for a3m_file in a3m_files:
        a3m_lines = get_a3m_lines(a3m_file)
    a3m_lines_list = ["".join(a3m_lines[n]) for n in Ms]

    if use_templates:
        templates = get_templates(
                x,
                Path(prefix),
                "pdb70.m8",
                num_templates,
            )

    return (a3m_lines_list, templates) if use_templates else a3m_lines_list


def run_mmseqs_command(mmseqs: Path, params: List[Union[str, Path]]):
    module = str(params[0])
    if module in MODULE_OUTPUT_POS:
        output_pos = MODULE_OUTPUT_POS[module]
        output_path = Path(params[output_pos]).with_suffix('.dbtype')
        if output_path.exists():
            logger.info(f"Skipping {module} because {output_path} already exists")
            return

    params_log = " ".join(str(i) for i in params)
    logger.info(f"Running {mmseqs} {params_log}")
    # hide MMseqs2 verbose paramters list that clogs up the log
    os.environ["MMSEQS_CALL_DEPTH"] = "1"
    subprocess.check_call([str(mmseqs)] + [str(i) for i in params])


# Lightly modified code from https://github.com/sokrypton/ColabFold
def run_local_mmseqs(
    x,
    base,
    use_env=True,
    use_templates=False,
    filter=0,
    num_templates=20,
    mmseqs_db=None,
    expand_eval: float = math.inf,
    align_eval: int = 10,
    diff: int = 3000,
    qsc: float = -20.0,
    max_accept: int = 1000000,
    prefilter_mode: int = 0,
    s: float = 8,
    db_load_mode: int = 2,
    threads: int = 32,
    gpu: int = 0,
    gpu_server: int = 0,
    unpack: bool = True,
) -> Sequence[object]:

    if filter:
        # 0.1 was not used in benchmarks due to POSIX shell bug in line above
        #  EXPAND_EVAL=0.1
        align_eval = 10
        qsc = 0.8
        max_accept = 100000

    mmseqs = Path("mmseqs")
    uniref_db = Path("uniref30_2302_db")
    metagenomic_db = Path("colabfold_envdb_202108_db")
    template_db = Path("pdb100_230517")

    base.mkdir(exist_ok=True, parents=True)
    query_file = base.joinpath("query.fas")
    with query_file.open("w") as f:
        query_seq_headername = 101
        f.write(f">{query_seq_headername}\n{x}\n")

    run_mmseqs_command(
        mmseqs,
        ["createdb", query_file, base.joinpath("qdb"), "--shuffle", "0"],
    )

    used_dbs = [uniref_db]
    if use_templates:
        used_dbs.append(template_db)
    if use_env:
        used_dbs.append(metagenomic_db)

    for db in used_dbs:
        if not mmseqs_db.joinpath(f"{db}.dbtype").is_file():
            raise FileNotFoundError(f"Database {db} does not exist")
        if (
            (
                not mmseqs_db.joinpath(f"{db}.idx").is_file()
                and not mmseqs_db.joinpath(f"{db}.idx.index").is_file()
            )
            or os.environ.get("MMSEQS_IGNORE_INDEX", False)
        ):
            logger.info("Search does not use index")
            db_load_mode = 0
            dbSuffix1 = "_seq"
            dbSuffix2 = "_aln"
            dbSuffix3 = ""
        else:
            dbSuffix1 = ".idx"
            dbSuffix2 = ".idx"
            dbSuffix3 = ".idx"

    search_param = ["--num-iterations", "3",
                    "--db-load-mode", str(db_load_mode),
                    "-a", "-e", "0.1", "--max-seqs", "10000"]
    if gpu:
        # gpu version only supports ungapped prefilter currently
        search_param += ["--gpu", str(gpu), "--prefilter-mode", "1"]
    else:
        search_param += ["--prefilter-mode", str(prefilter_mode)]
        # sensitivy can only be set for non-gpu version,
        # gpu version runs at max sensitivity
        if s is not None:
            search_param += ["-s", "{:.1f}".format(s)]
        else:
            search_param += ["--k-score", "'seq:96,prof:80'"]
    if gpu_server:
        search_param += ["--gpu-server", str(gpu_server)]

    filter_param = ["--filter-msa", str(filter),
                    "--filter-min-enable", "1000",
                    "--diff", str(diff),
                    "--qid", "0.0,0.2,0.4,0.6,0.8,1.0",
                    "--qsc", "0", "--max-seq-id", "0.95"]
    expand_param = ["--expansion-mode", "0",
                    "-e", str(expand_eval),
                    "--expand-filter-clusters", str(filter),
                    "--max-seq-id", "0.95"]

    if not base.joinpath("uniref.a3m").with_suffix('.a3m.dbtype').exists():
        run_mmseqs_command(mmseqs,
                           ["search", base.joinpath("qdb"),
                            mmseqs_db.joinpath(uniref_db),
                            base.joinpath("res"),
                            base.joinpath("tmp"),
                            "--threads", str(threads)] + search_param)
        run_mmseqs_command(mmseqs,
                           ["mvdb",
                            base.joinpath("tmp/latest/profile_1"),
                            base.joinpath("prof_res")])
        run_mmseqs_command(mmseqs,
                           ["lndb",
                            base.joinpath("qdb_h"),
                            base.joinpath("prof_res_h")])
        run_mmseqs_command(mmseqs,
                           ["expandaln",
                            base.joinpath("qdb"),
                            mmseqs_db.joinpath(f"{uniref_db}{dbSuffix1}"),
                            base.joinpath("res"),
                            mmseqs_db.joinpath(f"{uniref_db}{dbSuffix2}"),
                            base.joinpath("res_exp"),
                            "--db-load-mode", str(db_load_mode),
                            "--threads", str(threads)] + expand_param)
        run_mmseqs_command(mmseqs,
                           ["align",
                            base.joinpath("prof_res"),
                            mmseqs_db.joinpath(f"{uniref_db}{dbSuffix1}"),
                            base.joinpath("res_exp"),
                            base.joinpath("res_exp_realign"),
                            "--db-load-mode", str(db_load_mode),
                            "-e", str(align_eval),
                            "--max-accept", str(max_accept),
                            "--threads", str(threads),
                            "--alt-ali", "10", "-a"])
        run_mmseqs_command(mmseqs,
                           ["filterresult",
                            base.joinpath("qdb"),
                            mmseqs_db.joinpath(f"{uniref_db}{dbSuffix1}"),
                            base.joinpath("res_exp_realign"),
                            base.joinpath("res_exp_realign_filter"),
                            "--db-load-mode",
                            str(db_load_mode),
                            "--qid", "0",
                            "--qsc", str(qsc),
                            "--diff", "0",
                            "--threads", str(threads),
                            "--max-seq-id", "1.0",
                            "--filter-min-enable", "100"])
        run_mmseqs_command(mmseqs,
                           ["result2msa",
                            base.joinpath("qdb"),
                            mmseqs_db.joinpath(f"{uniref_db}{dbSuffix1}"),
                            base.joinpath("res_exp_realign_filter"),
                            base.joinpath("uniref.a3m"),
                            "--msa-format-mode", "6",
                            "--db-load-mode", str(db_load_mode),
                            "--threads", str(threads)] + filter_param)
        run_mmseqs_command(mmseqs, ["rmdb", base.joinpath("res_exp_realign_filter")])
        run_mmseqs_command(mmseqs, ["rmdb", base.joinpath("res_exp_realign")])
        run_mmseqs_command(mmseqs, ["rmdb", base.joinpath("res_exp")])
        run_mmseqs_command(mmseqs, ["rmdb", base.joinpath("res")])
    else:
        logger.info(f"Skipping {uniref_db} search because uniref.a3m already exists")

    bfd_exists = base.joinpath(
        "bfd.mgnify30.metaeuk30.smag30.a3m"
    ).with_suffix('.a3m.dbtype').exists()
    if use_env and not bfd_exists:
        run_mmseqs_command(mmseqs,
                           ["search",
                            base.joinpath("prof_res"),
                            mmseqs_db.joinpath(metagenomic_db),
                            base.joinpath("res_env"),
                            base.joinpath("tmp3"),
                            "--threads", str(threads)] + search_param)
        run_mmseqs_command(mmseqs,
                           ["expandaln",
                            base.joinpath("prof_res"),
                            mmseqs_db.joinpath(f"{metagenomic_db}{dbSuffix1}"),
                            base.joinpath("res_env"),
                            mmseqs_db.joinpath(f"{metagenomic_db}{dbSuffix2}"),
                            base.joinpath("res_env_exp"), "-e", str(expand_eval),
                            "--expansion-mode", "0",
                            "--db-load-mode", str(db_load_mode),
                            "--threads", str(threads)])
        run_mmseqs_command(mmseqs,
                           ["align",
                            base.joinpath("tmp3/latest/profile_1"),
                            mmseqs_db.joinpath(f"{metagenomic_db}{dbSuffix1}"),
                            base.joinpath("res_env_exp"),
                            base.joinpath("res_env_exp_realign"),
                            "--db-load-mode", str(db_load_mode),
                            "-e", str(align_eval),
                            "--max-accept", str(max_accept),
                            "--threads", str(threads),
                            "--alt-ali", "10", "-a"])
        run_mmseqs_command(mmseqs,
                           ["filterresult",
                            base.joinpath("qdb"),
                            mmseqs_db.joinpath(f"{metagenomic_db}{dbSuffix1}"),
                            base.joinpath("res_env_exp_realign"),
                            base.joinpath("res_env_exp_realign_filter"),
                            "--db-load-mode",
                            str(db_load_mode),
                            "--qid", "0",
                            "--qsc", str(qsc),
                            "--diff", "0",
                            "--max-seq-id", "1.0",
                            "--threads", str(threads),
                            "--filter-min-enable", "100"])
        run_mmseqs_command(mmseqs,
                           ["result2msa",
                            base.joinpath("qdb"),
                            mmseqs_db.joinpath(f"{metagenomic_db}{dbSuffix1}"),
                            base.joinpath("res_env_exp_realign_filter"),
                            base.joinpath("bfd.mgnify30.metaeuk30.smag30.a3m"),
                            "--msa-format-mode", "6",
                            "--db-load-mode", str(db_load_mode),
                            "--threads", str(threads)] + filter_param)
        run_mmseqs_command(mmseqs,
                           ["rmdb", base.joinpath("res_env_exp_realign_filter")])
        run_mmseqs_command(mmseqs, ["rmdb", base.joinpath("res_env_exp_realign")])
        run_mmseqs_command(mmseqs, ["rmdb", base.joinpath("res_env_exp")])
        run_mmseqs_command(mmseqs, ["rmdb", base.joinpath("res_env")])
    elif use_env:
        logger.info(
            f"Skipping {metagenomic_db} search because \
bfd.mgnify30.metaeuk30.smag30.a3m already exists")

    tmpl_db_exists = base.joinpath(
        f"{template_db}.m8"
    ).with_suffix('.m8.dbtype').exists()
    if use_templates and not tmpl_db_exists:
        run_mmseqs_command(mmseqs,
                           ["search",
                            base.joinpath("prof_res"),
                            mmseqs_db.joinpath(template_db),
                            base.joinpath("res_pdb"),
                            base.joinpath("tmp2"),
                            "--db-load-mode", str(db_load_mode),
                            "--threads", str(threads),
                            "-s", "7.5",
                            "-a", "-e", "0.1",
                            "--prefilter-mode", str(prefilter_mode)])
        run_mmseqs_command(mmseqs,
                           ["convertalis",
                            base.joinpath("prof_res"),
                            mmseqs_db.joinpath(f"{template_db}{dbSuffix3}"),
                            base.joinpath("res_pdb"),
                            base.joinpath(f"{template_db}"),
                            "--format-output",
                            "query,target,fident,alnlen,mismatch,\
gapopen,qstart,qend,tstart,tend,evalue,bits,cigar",
                            "--db-output", "1",
                            "--db-load-mode", str(db_load_mode),
                            "--threads", str(threads)])
        run_mmseqs_command(mmseqs, ["rmdb", base.joinpath("res_pdb")])
    elif use_templates:
        logger.info(
            f"Skipping {template_db} search because {template_db}.m8 already exists"
            )

    if use_env:
        run_mmseqs_command(mmseqs,
                           ["mergedbs", base.joinpath("qdb"),
                            base.joinpath("final.a3m"),
                            base.joinpath("uniref.a3m"),
                            base.joinpath("bfd.mgnify30.metaeuk30.smag30.a3m")])
        run_mmseqs_command(mmseqs,
                           ["rmdb",
                            base.joinpath("bfd.mgnify30.metaeuk30.smag30.a3m")])
        run_mmseqs_command(mmseqs,
                           ["rmdb",
                            base.joinpath("uniref.a3m")])
    else:
        run_mmseqs_command(mmseqs,
                           ["mvdb",
                            base.joinpath("uniref.a3m"),
                            base.joinpath("final.a3m")])
        run_mmseqs_command(mmseqs,
                           ["rmdb",
                            base.joinpath("uniref.a3m")])

    if unpack:
        run_mmseqs_command(mmseqs,
                           ["unpackdb",
                            base.joinpath("final.a3m"),
                            base.joinpath("."),
                            "--unpack-name-mode", "0",
                            "--unpack-suffix", ".a3m"])
        run_mmseqs_command(mmseqs,
                           ["rmdb",
                            base.joinpath("final.a3m")])

        if use_templates:
            run_mmseqs_command(mmseqs,
                               ["unpackdb",
                                base.joinpath(f"{template_db}"),
                                base.joinpath("."),
                                "--unpack-name-mode", "0",
                                "--unpack-suffix", ".m8"])
            if base.joinpath(f"{template_db}").exists():
                run_mmseqs_command(mmseqs, ["rmdb", base.joinpath(f"{template_db}")])

    run_mmseqs_command(mmseqs, ["rmdb", base.joinpath("prof_res")])
    run_mmseqs_command(mmseqs, ["rmdb", base.joinpath("prof_res_h")])
    shutil.rmtree(base.joinpath("tmp"))
    if use_templates:
        shutil.rmtree(base.joinpath("tmp2"))
    if use_env:
        shutil.rmtree(base.joinpath("tmp3"))

    if unpack:
        run_mmseqs_command(mmseqs, ["rmdb", base.joinpath("qdb")])
        run_mmseqs_command(mmseqs, ["rmdb", base.joinpath("qdb_h")])
        output_a3m = base.joinpath("0.a3m")
    else:
        output_a3m = base.joinpath("final.a3m")

    query_file.unlink()

    # Gather a3m lines in the same way as the API
    seqs = [x] if isinstance(x, str) else x
    N = 101
    seqs_unique = list(set(seqs))
    Ms = [N + seqs_unique.index(seq) for seq in seqs]
    a3m_lines = get_a3m_lines(output_a3m)
    a3m_lines_list = ["".join(a3m_lines[n]) for n in Ms]

    if use_templates:
        templates = get_templates(
            x,
            base,
            "0.m8",
            num_templates,
            mmseqs_db=mmseqs_db,
        )

    return (a3m_lines_list, templates) if use_templates else a3m_lines_list


def get_a3m_lines(output_a3m):
    a3m_lines: dict = {}
    update_M, M = True, None
    for line in open(output_a3m, "r"):
        if len(line) > 0:
            if "\x00" in line:
                line = line.replace("\x00", "")
                update_M = True
            if line.startswith(">") and update_M:
                M = int(line[1:].rstrip())
                update_M = False
                if M not in a3m_lines:
                    a3m_lines[M] = []
            a3m_lines[M].append(line)

    return a3m_lines


def get_templates(x, base, m8, num_templates, mmseqs_db=None):
    tested_pdbs = []
    templates = []
    logger.info("Finding and preparing templates")
    count = 0
    for line in open(base.joinpath(m8), "r"):
        template = {}
        if count < num_templates:
            p = line.rstrip().split()
            pdb, qid, alilen, tstart, tend = (
                p[1],
                float(p[2]),
                float(p[3]),
                int(p[8]),
                int(p[9]),
            )
            coverage = alilen / len(x)
            pdb_id = pdb.split("_")[0]

            # Use the same template filters as AF3 and only use 1 template per PDB
            if (
                qid == 1.0
                and coverage >= 0.95
                or coverage < 0.1
                or pdb_id in tested_pdbs
            ):
                continue

            pdb_id = pdb.split("_")[0]
            if mmseqs_db:
                cif_str = fetch_local_mmcif(pdb_id,
                                            pdb.split("_")[1],
                                            tstart,
                                            tend,
                                            base,
                                            mmseqs_db)
            else:
                cif_str = fetch_mmcif(pdb_id,
                                      pdb.split("_")[1],
                                      tstart,
                                      tend,
                                      base)
            template["mmcif"] = cif_str

            template_seq = extract_sequence_from_mmcif(StringIO(cif_str))
            query_indices, template_indices = align_and_map(x, template_seq)

            template["queryIndices"] = query_indices
            template["templateIndices"] = template_indices
            templates.append(template)
            tested_pdbs.append(pdb_id)
            count += 1
    logger.info(f"Found the following templates: {tested_pdbs}")
    return templates


def fetch_mmcif(
    pdb_id,
    chain_id,
    start,
    end,
    tmpdir,
):
    """
    Fetch the mmcif file for a given PDB ID
    and chain ID and prepare it for use in AlphaFold3
    """
    pdb_id = pdb_id.lower()
    url_base = "http://www.ebi.ac.uk/pdbe-srv/view/files/"
    url = url_base + pdb_id + ".cif"
    response = requests.get(url)
    text = response.text

    output = os.path.join(tmpdir, pdb_id + ".cif")
    with open(output, "w") as f:
        f.write(text)

    return get_mmcif(output, pdb_id, chain_id, start, end, tmpdir)


def fetch_local_mmcif(
        pdb_id,
        chain_id,
        start,
        end,
        tmpdir,
        mmseqs_db,
):
    """
    Fetch the mmcif file for a given PDB ID
    and chain ID and prepare it for use in AlphaFold3
    """
    pdb_id = pdb_id.lower()
    assert len(pdb_id) == 4, f"Invalid PDB ID: {pdb_id}"
    inner_code = pdb_id[1:3]
    mmcif_location = mmseqs_db.joinpath(f"pdb/divided/{inner_code}/{pdb_id}.cif.gz")
    if not mmcif_location.exists():
        raise FileNotFoundError(f"MMseqs2 database {mmcif_location} does not exist")
    with gzip.open(mmcif_location, "rb") as f:
        cif_str = f.read().decode("utf-8")

    output = os.path.join(tmpdir, pdb_id + ".cif")
    with open(output, "w") as f:
        f.write(cif_str)

    return get_mmcif(output, pdb_id, chain_id, start, end, tmpdir)


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="Add MMseqs2 unpaired MSA to AlphaFold3 json"
    )
    parser.add_argument("--input_json", help="Input alphafold3 json file")
    parser.add_argument("--output_json", help="Output alphafold3 json file")

    parser = mmseqs2_argparse_util(parser)
    parser = custom_template_argpase_util(parser)

    args = parser.parse_args()

    add_msa_to_json(  # pragma: no cover
        args.input_json,
        args.mmseqs_database,
        args.templates,
        args.num_templates,
        False,
        args.custom_template,
        args.custom_template_chain,
        args.target_id,
        output_json=args.output_json,
        to_file=True,
    )


if __name__ == "__main__":  # pragma: no cover
    main()



import pytest

from abcfold.scripts.ipsae import Ipsae


def test_af3_ipsae(output_objs):
    af3_file = output_objs.af3_output.cif_files["seed-1"][0]
    af3_pae_file = output_objs.af3_output.af3_pae_files["seed-1"][0]

    ipsae_calculator = Ipsae(
        af3_file,
        af3_pae_file,
        pae_cutoff=10.0,
        pae_format="alphafold3",
        distance_cutoff=5.0
    )
    scores = ipsae_calculator.main(verbose=False, output_csv=None)

    assert "pdockq" in scores
    assert "pdockq2" in scores
    assert "ligand_interface_scores" in scores
    assert "iptm_ipsae_scores" in scores

    assert scores["pdockq"]["A"]["B"] == pytest.approx(0.7226, rel=1e-3)
    assert scores["pdockq2"]["A"]["B"] == pytest.approx(0.733, rel=1e-3)
    assert scores["ligand_interface_scores"]["A"]["B"] == pytest.approx(
        0.5385, rel=1e-3
    )

    iptm = scores["iptm_ipsae_scores"]
    assert iptm["iptm_d0chn_asym"]["A"]["B"] == pytest.approx(
        0.8444376339, rel=1e-6
    )
    assert iptm["ipsae_d0res_asym"]["A"]["B"] == pytest.approx(
        0.7545828078, rel=1e-6
    )

    assert iptm["unique_residues_chain1"]["A"]["B"] == 391
    assert iptm["unique_residues_chain2"]["A"]["B"] == 393


def test_boltz_ipsae(output_objs):
    boltz_file = output_objs.boltz_output.cif_files["seed-1"][0]
    boltz_pae_file = output_objs.boltz_output.af3_pae_files["seed-1"][0]

    ipsae_calculator = Ipsae(
        boltz_file,
        boltz_pae_file,
        pae_cutoff=10.0,
        pae_format="alphafold3",
        distance_cutoff=5.0
    )
    scores = ipsae_calculator.main(verbose=False, output_csv=None)

    assert "pdockq" in scores
    assert "pdockq2" in scores
    assert "ligand_interface_scores" in scores
    assert "iptm_ipsae_scores" in scores

    assert scores["pdockq"]["A"]["B"] == pytest.approx(0.7248, rel=1e-3)
    assert scores["pdockq2"]["A"]["B"] == pytest.approx(0.5144, rel=1e-3)
    assert scores["ligand_interface_scores"]["A"]["B"] == pytest.approx(
        0.7273, rel=1e-3
    )

    iptm = scores["iptm_ipsae_scores"]
    assert iptm["iptm_d0chn_asym"]["A"]["B"] == pytest.approx(
        0.9633755758902331, rel=1e-6
    )
    assert iptm["ipsae_d0res_asym"]["A"]["B"] == pytest.approx(
        0.945514520263416, rel=1e-6
    )

    assert iptm["unique_residues_chain1"]["A"]["B"] == 389
    assert iptm["unique_residues_chain2"]["A"]["B"] == 390


def test_chai_ipsae(output_objs):
    chai_file = output_objs.chai_output.cif_files["seed-1"][0]
    chai_pae_file = output_objs.chai_output.af3_pae_files["seed-1"][0]

    ipsae_calculator = Ipsae(
        chai_file,
        chai_pae_file,
        pae_cutoff=10.0,
        pae_format="alphafold3",
        distance_cutoff=5.0
    )
    scores = ipsae_calculator.main(verbose=False, output_csv=None)

    assert "pdockq" in scores
    assert "pdockq2" in scores
    assert "ligand_interface_scores" in scores
    assert "iptm_ipsae_scores" in scores

    assert scores["pdockq"]["A"]["B"] == pytest.approx(0.721, rel=1e-3)
    assert scores["pdockq2"]["A"]["B"] == pytest.approx(0.8039, rel=1e-3)
    assert scores["ligand_interface_scores"]["A"]["B"] == pytest.approx(
        0.7401, rel=1e-3
    )

    iptm = scores["iptm_ipsae_scores"]
    assert iptm["iptm_d0chn_asym"]["A"]["B"] == pytest.approx(
        0.9797472978285606, rel=1e-6
    )
    assert iptm["ipsae_d0res_asym"]["A"]["B"] == pytest.approx(
        0.9679382445215664, rel=1e-6
    )

    assert iptm["unique_residues_chain1"]["A"]["B"] == 391
    assert iptm["unique_residues_chain2"]["A"]["B"] == 393


def test_protenix_ipsae(output_objs):
    protenix_file = output_objs.protenix_output.cif_files["seed-1"][0]
    protenix_pae_file = output_objs.protenix_output.af3_pae_files["seed-1"][0]

    ipsae_calculator = Ipsae(
        protenix_file,
        protenix_pae_file,
        pae_cutoff=10.0,
        pae_format="alphafold3",
        distance_cutoff=5.0
    )
    scores = ipsae_calculator.main(verbose=False, output_csv=None)

    assert "pdockq" in scores
    assert "pdockq2" in scores
    assert "ligand_interface_scores" in scores
    assert "iptm_ipsae_scores" in scores

    assert scores["pdockq"]["A"]["B"] == pytest.approx(0.7202, rel=1e-3)
    assert scores["pdockq2"]["A"]["B"] == pytest.approx(0.7041, rel=1e-3)
    assert scores["ligand_interface_scores"]["A"]["B"] == pytest.approx(
        0.6987, rel=1e-3
    )

    iptm = scores["iptm_ipsae_scores"]
    assert iptm["iptm_d0chn_asym"]["A"]["B"] == pytest.approx(
        0.9689027305100429, rel=1e-6
    )
    assert iptm["ipsae_d0res_asym"]["A"]["B"] == pytest.approx(
        0.9517480877376054, rel=1e-6
    )

    assert iptm["unique_residues_chain1"]["A"]["B"] == 391
    assert iptm["unique_residues_chain2"]["A"]["B"] == 391



import json
import tempfile
from pathlib import Path

from abcfold.protenix.af3_to_protenix import ProtenixJson

# flake8: noqa


def test_af3_to_protenix(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        protenix_json = ProtenixJson(temp_dir)

        data = protenix_json.json_to_json(test_data.test_inputAB_json)

        reference = {
            'name': '2PV7',
            'modelSeeds': [1],
            'sequences':
                [
                    {'proteinChain': {'sequence': 'GMRES', 'count': 2}},
                    {'proteinChain': {'sequence': 'YANEN', 'count': 1}},
                    {'ligand': {'ligand': 'CCD_ATP', 'count': 2}},
                    {'ligand': {'ligand': 'CC(=O)OC1C[NH+]2CCC1CC2', 'count': 1}}
                ]
        }

        assert data == reference


def test_af3_to_protenix_rna(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        protenix_json = ProtenixJson(temp_dir)

        data = protenix_json.json_to_json(test_data.test_inputRNA_json)

        reference = {
            'name': 'RNA_example',
            'modelSeeds': [1],
            'sequences':
                [
                    {'rnaSequence': {'sequence': 'AGCU', 'count': 1}}
                ]
        }

        assert data == reference


def test_af3_to_protenix_dna(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        protenix_json = ProtenixJson(temp_dir)

        data = protenix_json.json_to_json(test_data.test_inputDNA_json)

        reference = {
            'name': 'DNA_example',
            'modelSeeds': [1],
            'sequences':
                [
                    {'dnaSequence': {'sequence': 'AGCT', 'count': 2}}
                ]
        }

        assert data == reference


def test_af3_to_protenix_ligand(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        protenix_json = ProtenixJson(temp_dir)

        data = protenix_json.json_to_json(test_data.test_inputLIG_json)

        reference = {
            'name': '2PV7',
            'modelSeeds': [1],
            'sequences':
                [
                    {'proteinChain': {'sequence': 'GMRESYANENQFGFKTINSDIHKIVIVGGYGKLGGLFARYLRASGYPISILDREDWAVAESILANADVVIVSVPINLTLETIERLKPYLTENMLLADLTSVKREPLAKMLEVHTGAVLGLHPMFGADIASMAKQVVVRCDGRFPERYEWLLEQIQIWGAKIYQTNATEHDHNMTYIQALRHFSTFANGLHLSKQPINLANLLALSSPIYRLELAMIGRLFAQDAELYADIIMDKSENLAVIETLKQTYDEALTFFENNDRQGFIDAFHKVRDWFGDYSEQFLKESRQLLQQANDLKQG', 'count': 2}},
                    {'ligand': {'ligand': 'CCD_ATP', 'count': 2}},
                    {'ligand': {'ligand': 'CC(=O)OC1C[NH+]2CCC1CC2', 'count': 1}},
                    {'ligand': {'ligand': 'CCCCCCCCCCCC(O)=O', 'count': 2}},
                    {'ion': {'ion': 'CCD_MG', 'count': 1}}
                ]
            }

        assert data == reference


def test_af3_to_protenix_ptm(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        protenix_json = ProtenixJson(temp_dir)

        data = protenix_json.json_to_json(test_data.test_inputPTM_json)

        reference = {
            'name': 'PTM example',
            'modelSeeds': [1],
            'sequences':
                [
                    {'proteinChain': {'sequence': 'PVLSCGEWQL',
                                      'count': 1,
                                      'modifications': [
                                          {'ptmType': 'CCD_HY3', 'ptmPosition': 1},
                                          {'ptmType': 'CCD_P1L', 'ptmPosition': 5}
                                      ]
                                     }
                    },
                    {'rnaSequence': {'sequence': 'AGCU',
                                     'count': 1,
                                     'modifications': [
                                         {'modificationType': 'CCD_2MG',
                                          'basePosition': 1},
                                         {'modificationType': 'CCD_5MC',
                                          'basePosition': 4}
                                     ]
                                    }
                    }
                ]
        }

        assert data == reference

def test_af3_to_protenix_constraints(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        protenix_json = ProtenixJson(temp_dir)

        data = protenix_json.json_to_json(test_data.test_inputBOND_json)

        reference = {
            'name': '7SYZ',
            'modelSeeds': [1],
            'sequences': [
                {'proteinChain': {'sequence': 'MMADSKLVSLNNNLSGKIKDQGKVIKNYYGTMDIKKINDGLLDSKILGAFNTVIALLGSIIIIVMNIMIIQNYTRTTDNQALIKESLQSVQQQIKALTDKIGTEIGPKVSLIDTSSTITIPANIGLLGSKISQSTSSINENVNDKCKFTLPPLKIHECNISCPNPLPFREYRPISQGVSDLVGLPNQICLQKTTSTILKPRLISYTLPINTREGVCITDPLLAVDNGFFAYSHLEKIGSCTRGIAKQRIIGVGEVLDRGDKVPSMFMTNVWTPPNPSTIHHCSSTYHEDFYYTLCAVSHVGDPILNSTSWTESLSLIRLAVRPKSDSGDYNQKYIAITKVERGKYDKVMPYGPSGIKQGDTLYFPAVGFLPRTEFQYNDSNCPIIHCKYSKAENCRLSMGVNSKSHYILRSGLLKYNLSLGGDIILQFIEIADNRLTIGSPSKIYNSLGQPVFYQASYSWDTMIKLGDVDTVDPLRVQWRNNSVISRPGQSQCPRFNVCPEVCWEGTYNDAFLIDRLNWVSAGVYLNSNQTAENPVFAVFKDNEILYQVPLAEDDTNAQKTITDCFLLENVIWCISLVEIYDTGDSVIRPKLFAVKIPAQCSES', 'count': 1}},
                {'proteinChain': {'sequence': 'QIQLVQSGPELKKPGETVKISCTTSGYTFTNYGLNWVKQAPGKGFKWMAWINTYTGEPTYADDFKGRFAFSLETSASTTYLQINNLKNEDMSTYFCARSGYYDGLKAMDYWGQGTSVTVSSAKTTPPSVYPLAPGSAAQTNSMVTLGCLVKGYFPEPVTVTWNSGSLSSGVHTFPAVLQSDLYTLSSSVTVPSSTWPSETVTCNVAHPASSTKVDKKIVPRDC', 'count': 1}},
                {'proteinChain': {'sequence': 'DVLMIQTPLSLPVSLGDQASISCRSSQSLIHINGNTYLEWYLQKPGQSPKLLIYKVSNRFSGVPDRFSGSGSGTDFTLKISRVEAEDLGVYYCFQGSHVPFTFGAGTKLELKRADAAPTVSIFPPSSEQLTSGGASVVCFLNNFYPKDINVKWKIDGSERQNGVLNSWTDQDSKDSTYSMSSTLTLTKDEYERHNSYTCEATHKTSTSPIVKSFNRNECVY', 'count': 1}}
            ],
            "contact": [
                {
                    "entity1": 1,
                    "copy1": 1,
                    "position1": 387,
                    "atom1": "CA",
                    "entity2": 2,
                    "copy2": 1,
                    "position2": 101,
                    "atom2": "CA",
                    "max_distance": 6,
                    "min_distance": 0
                },
                {
                    "entity1": 3,
                    "copy1": 1,
                    "position1": 32,
                    "atom1": "CA",
                    "entity2": 1,
                    "copy2": 1,
                    "atom2": "CA",
                    "position2": 483,
                    "max_distance": 6,
                    "min_distance": 0
                }
            ]
        }

        assert data == reference

def test_protenix_output_msa(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        protenix_json = ProtenixJson(temp_dir)

        data = protenix_json.json_to_json(test_data.test_inputAmsa_json)
        msa_path = (
            data["sequences"][0]["proteinChain"]
            .get("msa", {})
            .get("precomputed_msa_dir")
        )
        # MSA directory has a random path, so just check that it exists then give
        # it a placeholder value for comparison
        assert msa_path is not None
        assert Path(msa_path).exists()
        data["sequences"][0]["proteinChain"]["msa"]["precomputed_msa_dir"] = (
            "PRECOMPUTED_MSA_DIR"
        )
        print(data)

        reference = {
            'name': '2PV7',
            'modelSeeds': [1],
            'contact': [
                {'entity1': 1,
                 'position1': 1,
                 'copy1': 1, 'atom1':
                 'CA', 'entity2': 1,
                 'position2': 20,
                 'copy2': 1,
                 'atom2': 'CA',
                 'max_distance': 6,
                 'min_distance': 0}
            ],
            'sequences': [
                {'proteinChain': {'sequence': 'GMRESYANENQFGFKTINSDIHKIVIVGGYGKLGGLFARYLRASGYPISILDREDWAVAESILANADVVIVSVPINLTLETIERLKPYLTENMLLADLTSVKREPLAKMLEVHTGAVLGLHPMFGADIASMAKQVVVRCDGRFPERYEWLLEQIQIWGAKIYQTNATEHDHNMTYIQALRHFSTFANGLHLSKQPINLANLLALSSPIYRLELAMIGRLFAQDAELYADIIMDKSENLAVIETLKQTYDEALTFFENNDRQGFIDAFHKVRDWFGDYSEQFLKESRQLLQQANDLKQG',
                                  'count': 1,
                                  'msa': {'precomputed_msa_dir': 'PRECOMPUTED_MSA_DIR',
                                          'pairing_db': 'uniref100'
                                          }
                                }
                }
            ]
        }

        assert data == reference

def test_protenix_write_json(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        protenix_json = ProtenixJson(temp_dir)

        protenix_json.json_to_json(test_data.test_inputAB_json)

        out_file = Path(temp_dir) / "protenix_output.json"
        protenix_json.write_json(out_file)

        reference = [
            {
            'name': '2PV7',
            'modelSeeds': [1],
            'sequences':
                [
                    {'proteinChain': {'sequence': 'GMRES', 'count': 2}},
                    {'proteinChain': {'sequence': 'YANEN', 'count': 1}},
                    {'ligand': {'ligand': 'CCD_ATP', 'count': 2}},
                    {'ligand': {'ligand': 'CC(=O)OC1C[NH+]2CCC1CC2', 'count': 1}}
                ]
            }
        ]

        with open(out_file, "r") as f:
            written_data = f.read()
        written_data = json.loads(written_data)


        assert written_data == reference



import json
import logging
import random
import string
from pathlib import Path
from typing import Any, Dict, List, Union

logger = logging.getLogger("logger")


ION_CCD_CODES = [
    # alkali / alkaline earth
    "LI", "NA", "K", "RB", "CS",
    "MG", "CA", "SR", "BA",
    # transition metals
    "ZN", "FE", "FE2", "FE3", "CU", "MN", "CO", "NI",
    "CD", "HG",
    # halides
    "F", "CL", "BR", "I",
]


class ProtenixJson:
    """
    Object to convert an AlphaFold3 json file to a Protenix JSON file.
    """

    def __init__(self, working_dir: Union[str, Path], create_files: bool = True):
        self.working_dir = working_dir
        self.seeds: list = [42]
        self.__ids: Dict = {}
        self.__id_counter: int = 1
        self.__create_files = create_files
        self.protenix_dict: Dict = {}

    @property
    def chain_ids(self) -> Dict:
        return self.__ids

    def msa_to_file(self, msa: str, file_path: Union[str, Path]):
        """
        Takes a msa string and writes it to a file

        Args:
            msa (str): msa string
            file_path (Union[str, Path]): file path to write the msa to

        Returns:
            None
        """

        with open(file_path, "w") as f:
            f.write(msa)

    def json_to_json(
        self,
        json_file_or_dict: Union[dict, str, Path],
    ):
        """
        Main function to convert an AF3 json file or dict to a Protenix json string

        Args:
            json_file_or_dict (Union[dict, str, Path]): json file or dict

        Returns:
            Dict: protenix dictionary
        """
        logger.info("Converting input json to a Protenix compatible json file")
        if isinstance(json_file_or_dict, str) or isinstance(json_file_or_dict, Path):
            with open(json_file_or_dict, "r") as f:
                json_dict = json.load(f)
        else:
            json_dict = json_file_or_dict

        protenix_sequences = []
        for key, value in json_dict.items():
            if key == "name":
                self.protenix_dict["name"] = value
            if key == "modelSeeds":
                if isinstance(value, list):
                    self.seeds = value
                elif isinstance(value, int):
                    self.seeds = [value]
                self.protenix_dict["modelSeeds"] = self.seeds

            if key == "sequences":
                for entry in value:
                    if "protein" in entry:
                        protein_entry = self.convert_protein(entry["protein"])
                        protenix_sequences.append(protein_entry)
                    elif "rna" in entry:
                        rna_entry = self.convert_rna(entry["rna"])
                        protenix_sequences.append(rna_entry)
                    elif "dna" in entry:
                        dna_entry = self.convert_dna(entry["dna"])
                        protenix_sequences.append(dna_entry)
                    elif "ligand" in entry:
                        ligand_entry = self.convert_ligand(entry["ligand"])
                        protenix_sequences.append(ligand_entry)

            if key == "bondedAtomPairs" and value is not None:
                contact = self.convert_bonded_atom_pairs(value)
                self.protenix_dict["contact"] = contact

        self.protenix_dict["sequences"] = protenix_sequences

        return self.protenix_dict

    def convert_protein(self, seq_dict) -> Dict[str, Any]:
        sequence = seq_dict["sequence"]
        chain_ids = seq_dict.get("id", [])
        for chain in chain_ids:
            self.__ids[chain] = self.__id_counter
            self.__id_counter += 1
        count = len(chain_ids) if chain_ids else 1

        protein_chain = {
            "sequence": sequence,
            "count": count,
        }

        modifications = seq_dict.get("modifications")
        if modifications:
            prefixed = []
            for mod in modifications:
                m = dict(mod)
                typ = str(m.get("ptmType", ""))
                if not typ.startswith("CCD_"):
                    typ = "CCD_" + typ
                m["ptmType"] = typ
                prefixed.append(m)
            protein_chain["modifications"] = prefixed

        unpaired_msa = seq_dict.get("unpairedMsa")
        paired_msa = seq_dict.get("pairedMsa")
        random_string = ''.join(random.choices(string.ascii_letters, k=5))
        msa_dir = Path(self.working_dir) / random_string
        if unpaired_msa and self.__create_files:
            if not msa_dir.exists():
                msa_dir.mkdir(parents=True, exist_ok=True)
            self.msa_to_file(
                unpaired_msa,
                msa_dir / "non_pairing.a3m"
            )

        if paired_msa and self.__create_files:
            if not msa_dir.exists():
                msa_dir.mkdir(parents=True, exist_ok=True)
            self.msa_to_file(
                paired_msa,
                msa_dir / "pairing.a3m"
            )

        if unpaired_msa or paired_msa and self.__create_files:
            protein_chain["msa"] = {
                "precomputed_msa_dir": msa_dir.as_posix(),
                "pairing_db": "uniref100"
            }

        return {
            "proteinChain": protein_chain
        }

    def convert_rna(self, seq_dict) -> Dict[str, Any]:
        sequence = seq_dict["sequence"]
        chain_ids = seq_dict.get("id", [])
        for chain in chain_ids:
            self.__ids[chain] = self.__id_counter
            self.__id_counter += 1
        count = len(chain_ids) if chain_ids else 1

        rna_chain = {
            "sequence": sequence,
            "count": count,
        }

        modifications = seq_dict.get("modifications")
        if modifications:
            prefixed = []
            for mod in modifications:
                m = dict(mod)
                typ = str(m.get("modificationType", ""))
                if not typ.startswith("CCD_"):
                    typ = "CCD_" + typ
                m["modificationType"] = typ
                prefixed.append(m)
            rna_chain["modifications"] = prefixed

        return {
            "rnaSequence": rna_chain
        }

    def convert_dna(self, seq_dict) -> Dict[str, Any]:
        sequence = seq_dict["sequence"]
        chain_ids = seq_dict.get("id", [])
        for chain in chain_ids:
            self.__ids[chain] = self.__id_counter
            self.__id_counter += 1
        count = len(chain_ids) if chain_ids else 1

        dna_chain = {
            "sequence": sequence,
            "count": count,
        }

        modifications = seq_dict.get("modifications")
        if modifications:
            prefixed = []
            for mod in modifications:
                m = dict(mod)
                typ = str(m.get("modificationType", ""))
                if not typ.startswith("CCD_"):
                    typ = "CCD_" + typ
                m["modificationType"] = typ
                prefixed.append(m)
            dna_chain["modifications"] = prefixed

        return {
            "dnaSequence": dna_chain
        }

    def convert_ligand(self, seq_dict) -> Dict[str, Any]:
        chain_ids = seq_dict.get("id", [])
        for chain in chain_ids:
            self.__ids[chain] = self.__id_counter
            self.__id_counter += 1
        count = len(chain_ids) if chain_ids else 1

        if "ccdCodes" in seq_dict:
            ligand_id = seq_dict["ccdCodes"][0]
            ligand = f"CCD_{ligand_id}"
        else:
            ligand = seq_dict["smiles"]
            ligand_id = ligand

        if ligand_id.upper() in ION_CCD_CODES:
            ligand_chain = {
                "ion": ligand,
                "count": count,
            }
            return {
                "ion": ligand_chain
            }
        else:
            ligand_chain = {
                "ligand": ligand,
                "count": count,
            }
            return {
                "ligand": ligand_chain
            }

    def convert_bonded_atom_pairs(self,
                                  bonded_atom_pairs: List[List[List[Union[str, int]]]]):
        contacts = []
        for pair in bonded_atom_pairs:
            entity1, position1, atom1 = pair[0]
            entity2, position2, atom2 = pair[1]
            contact = {
                "entity1": self.__ids[entity1],
                "position1": position1,
                "copy1": 1,
                "atom1": atom1,
                "entity2": self.__ids[entity2],
                "position2": position2,
                "copy2": 1,
                "atom2": atom2,
                "max_distance": 6,
                "min_distance": 0
            }
            contacts.append(contact)
        return contacts

    def write_json(self, out_file: Union[str, Path]):
        """
        Write the Protenix json to a file

        Args:
            out_file (Union[str, Path]): output file path

        Returns:
            None
        """

        with open(out_file, "w") as f:
            json.dump([self.protenix_dict], f, indent=4)



#!/usr/bin/env python

import copy
import logging
from collections import defaultdict
from pathlib import Path

import numpy as np
import pandas as pd
from Bio.PDB.Polypeptide import is_aa
from Bio.SeqUtils import seq1

from abcfold.output.file_handlers import (CifFile, ConfidenceJsonFile,
                                          FileTypes, NpyFile, NpzFile, PklFile,
                                          ResidueCountType)
from abcfold.output.utils import Af3Pae

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s"
)
logger = logging.getLogger("logger")


class Ipsae():
    def __init__(self,
                 input_model,
                 pae_file,
                 pae_cutoff=5,
                 pae_format="alphafold3",
                 distance_cutoff=10):
        self.input_model = input_model
        self.pae_file = pae_file
        self.pae_cutoff = pae_cutoff
        self.pae_format = pae_format
        self.distance_cutoff = distance_cutoff

        # Read structure + PAE data
        if isinstance(self.input_model, CifFile):
            self.struct = copy.deepcopy(self.input_model)
        else:
            self.struct = CifFile(self.input_model)
        self.process_input()

    @staticmethod
    def ptm_func(x, d0):
        x = np.asarray(x)
        d0 = float(d0)
        return 1.0 / (1.0 + np.square(x / d0))

    @staticmethod
    def calc_d0(L, pair_type='protein'):
        """
        Vectorized d0 calculator.

        - L : scalar or array-like of sequence lengths. Values < 27 are treated as 27.
        - pair_type : string. 'nucleic_acid' (case-insensitive) uses minimum d0 = 2.0;
                    any other value uses minimum d0 = 1.0.

        Returns a NumPy array (or scalar ndarray) of d0 values.
        """
        if not isinstance(pair_type, str):
            raise TypeError("pair_type must be a string")

        L = np.asarray(L, dtype=float)
        L = np.maximum(L, 27.0)

        min_value = 2.0 if pair_type.lower() == 'nucleic_acid' else 1.0
        d0 = 1.24 * np.power(L - 15.0, 1.0 / 3.0) - 1.8
        return np.maximum(min_value, d0)

    @staticmethod
    def nested_chain_map(default, chain_ids):
        return {c1: {c2: default() for c2 in chain_ids if c1 != c2} for c1 in chain_ids}

    @staticmethod
    def pdockq(plddt_mean, n_pairs):
        x = plddt_mean * np.log10(n_pairs)
        return 0.724 / (1 + np.exp(-0.052*(x - 152.611))) + 0.018

    @staticmethod
    def pdockq2(plddt_mean, mean_ptm):
        x2 = plddt_mean * mean_ptm
        return 1.31 / (1 + np.exp(-0.075*(x2 - 84.733))) + 0.005

    @property
    def chain_res_map(self):
        if not hasattr(self, '_chain_res_map'):
            self._chain_res_map = np.asarray([
                chain.id
                for chain in self.struct.get_chains()
                for res in chain if is_aa(res.resname, standard=False)
            ], dtype=object)
        return self._chain_res_map

    @property
    def nres(self):
        if not hasattr(self, '_nres'):
            self._nres = len(self.chain_res_map)
        return self._nres

    @property
    def chain_ids(self):
        if not hasattr(self, '_chain_ids'):
            self._chain_ids = sorted(set(self.chain_res_map))
        return self._chain_ids

    def process_input(self):
        """
        Process input files and setup data types
        """

        if isinstance(self.pae_file, ConfidenceJsonFile):
            file_ = self.pae_file
        else:
            # Find PAE data format
            suffix = Path(self.pae_file).suffix[1:]
            if suffix == FileTypes.NPZ.value:
                file_ = NpzFile(self.pae_file)
            elif suffix == FileTypes.NPY.value:
                file_ = NpyFile(self.pae_file)
            elif suffix == FileTypes.JSON.value:
                file_ = ConfidenceJsonFile(self.pae_file)
            elif suffix == FileTypes.PKL.value:
                file_ = PklFile(self.pae_file)
            else:
                raise ValueError(f"Unsupported PAE file type: {suffix}")

        # Construct input_params obj dict with protein seqs to avoid error msg
        self.struct.input_params = {"sequences": []}
        for chain in self.struct.get_chains():
            seq_data = {}
            seq_data["protein"] = {}
            if all([is_aa(res.resname) for res in chain.get_residues()]):
                seq_data["protein"]["ID"] = [chain.id]
                seq_data["protein"]["sequence"] = "".join(
                        [seq1(residue.get_resname()) for residue in chain]
                    )
                self.struct.input_params["sequences"].append(seq_data)

        # Get PAE data for different formats
        if self.pae_format == "alphafold2":
            self.pae_data = Af3Pae.from_alphafold2(
                file_.data,
                self.struct,
            ).scores
        elif self.pae_format == "alphafold3":
            self.pae_data = Af3Pae.from_alphafold3(
                file_.data,
                self.struct,
            ).scores
        elif self.pae_format == "boltz":
            self.pae_data = Af3Pae.from_boltz(
                file_.data,
                self.struct,
            ).scores
        elif self.pae_format == "chai":
            self.pae_data = Af3Pae.from_chai1(
                file_.data,
                self.struct,
            ).scores
        elif self.pae_format == "colabfold":
            self.pae_data = Af3Pae.from_colabfold(
                file_.data,
                self.struct,
            ).scores
        elif self.pae_format == "protenix":
            self.pae_data = Af3Pae.from_protenix(
                file_.data,
                self.struct,
            ).scores
        else:
            raise ValueError(f"Unsupported PAE format: {self.pae_format}")

        # Extract pLDDTs
        if "plddt" in self.pae_data:
            self.plddt = np.array(self.pae_data["plddt"])
        else:
            plddt = self.struct.get_plddt_per_residue(
                method=ResidueCountType.IPSAE.value
            )
            self.plddt = np.array(
                sum([plddt[chain] for chain in plddt.keys()
                     if not self.struct.check_ligand(chain)
                     ], [])
            )

        # Generate PAE matrix
        if 'pae' in self.pae_data:
            self.pae_matrix = np.array(self.pae_data['pae'])
        elif 'predicted_aligned_error' in self.pae_data:
            self.pae_matrix = np.array(self.pae_data['predicted_aligned_error'])
        else:
            self.pae_matrix = None
        if self.pae_format in ["alphafold3", "boltz", "chai"]:
            token_mask = self.get_token_mask()
            self.pae_matrix = self.pae_matrix[np.ix_(token_mask, token_mask)]

        return

    def get_token_mask(self):
        # Mask out non-standard atoms possible
        token_mask = []
        for chain in self.struct.get_chains():
            for residue in chain:
                if is_aa(residue.resname, standard=True):
                    if any((atom.id == "CA") or ("C1" in atom.id) for atom in residue):
                        token_mask.append(1)
                elif is_aa(residue.resname, standard=False):
                    for atom in residue:
                        if (atom.id == "CA") or ("C1" in atom.id):
                            token_mask.append(1)
                        else:
                            token_mask.append(0)
                elif self.struct.check_ligand(chain):
                    for atom in residue:
                        token_mask.append(0)
                else:
                    for atom in residue:
                        token_mask.append(0)
        return np.array(token_mask, dtype=bool)

    def get_cb_distance(self):
        cb_coords = []
        for chain in self.struct.get_chains():
            for residue in chain:
                chosen_atom = None
                for atom in residue:
                    if not is_aa(residue.resname, standard=False):
                        continue
                    if atom.id == "CB":
                        chosen_atom = atom
                        break
                    if "C3" in atom.id:
                        chosen_atom = atom
                        break
                    if residue.resname == "GLY" and atom.id == "CA":
                        chosen_atom = atom
                        break

                if chosen_atom:
                    cb_coords.append(chosen_atom.get_coord())

        cb_coords = np.array([x for x in cb_coords])
        distances = np.sqrt(
            ((cb_coords[:, None, :] - cb_coords[None, :, :])**2).sum(axis=2)
        )
        return distances

    def compute_pdockq(self, distances, cutoff=8.0):
        """
        Compute pDockQ and pDockQ2 for all inter-chain interfaces.

        Parameters
        ----------
        distances : (N, N) ndarray
            Matrix of residue–residue Cβ distances.
        cutoff : float
            Distance threshold for residue–residue contacts.

        Returns
        -------
        pdockq_scores, pdockq2_scores : dict[chain1][chain2] → float
        """

        if distances.shape[0] != self.nres or distances.shape[1] != self.nres:
            raise ValueError(
                f"Residue count mismatch: distances.shape={distances.shape}, \
                residues={self.nres}"
            )

        pdockq_scores = self.nested_chain_map(lambda: 0.0, self.chain_ids)
        pdockq2_scores = self.nested_chain_map(lambda: 0.0, self.chain_ids)
        unique_residues = self.nested_chain_map(set, self.chain_ids)

        pair_counts = self.nested_chain_map(lambda: 0, self.chain_ids)
        pae_ptm_sums = self.nested_chain_map(lambda: 0.0, self.chain_ids)

        for i in range(self.nres):
            c_i = self.chain_res_map[i]

            pc = pair_counts[c_i]
            ur = unique_residues[c_i]
            pps = pae_ptm_sums[c_i]

            close_js = np.where(distances[i] <= cutoff)[0]
            if close_js.size == 0:
                continue

            per_chain_js = defaultdict(list)
            for j in close_js:
                if j == i:
                    continue
                c_j = self.chain_res_map[j]
                if c_j == c_i:
                    continue
                per_chain_js[c_j].append(j)

            for c_j, js in per_chain_js.items():
                js_arr = np.asarray(js, dtype=int)

                n_pairs_here = js_arr.size

                pc[c_j] += n_pairs_here
                ur[c_j].add(i)
                ur[c_j].update(js_arr)

                pae_vals = self.pae_matrix[i, js_arr]
                ptm_vals = self.ptm_func(pae_vals, 10.0)
                pps[c_j] += ptm_vals.sum()

        for c1 in self.chain_ids:
            for c2 in self.chain_ids:
                if c1 == c2:
                    continue

                n_pairs = pair_counts[c1][c2]
                if n_pairs == 0:
                    pdockq_scores[c1][c2] = 0.0
                    pdockq2_scores[c1][c2] = 0.0
                    continue

                uniq_res_list = sorted(unique_residues[c1][c2])
                mean_plddt = float(
                    np.mean(self.plddt[uniq_res_list])
                    ) if len(uniq_res_list) > 0 else 0.0

                pdq = self.pdockq(mean_plddt, n_pairs)
                pdockq_scores[c1][c2] = float(np.round(pdq, 4))

                mean_ptm = pae_ptm_sums[c1][c2] / n_pairs
                pdq2 = self.pdockq2(mean_plddt, mean_ptm)
                pdockq2_scores[c1][c2] = float(np.round(pdq2, 4))
        return pdockq_scores, pdockq2_scores

    def compute_ligand_interface_scores(self):
        """
        Compute ligand interface scores (LIS) between all chain pairs
        using the PAE matrix and a PAE cutoff stored in self.pae_cutoff.
        """

        LIS = self.nested_chain_map(lambda: 0.0, self.chain_ids)

        for chain1 in self.chain_ids:
            for chain2 in self.chain_ids:
                if chain1 == chain2:
                    continue

                # Mask residues from chain1 vs chain2
                mask = (self.chain_res_map[:, None] == chain1) & \
                    (self.chain_res_map[None, :] == chain2)
                selected_pae = self.pae_matrix[mask]

                if selected_pae.size == 0:
                    LIS[chain1][chain2] = 0.0
                    continue

                # Apply cutoff
                valid_pae = selected_pae[selected_pae <= 12]
                if valid_pae.size == 0:
                    LIS[chain1][chain2] = 0.0
                    continue

                # Score = (cutoff - pae) / cutoff
                scores = (12 - valid_pae) / 12
                LIS[chain1][chain2] = float(np.round(np.mean(scores), 4))

        return LIS

    def compute_iptm_ipsae(self, distances):
        """
        Compute iPTM and IPSAE scores for all inter-chain interfaces.

        Parameters
        ----------
        distances : (N, N) ndarray
            Matrix of residue–residue Cβ distances.

        Returns
        -------
        dicts containing per-residue and per-chain-pair scores:
            iptm_d0chn_byres, ipsae_d0chn_byres,
            ipsae_d0dom_byres, ipsae_d0res_byres,
            iptm_d0chn_asym, ipsae_d0chn_asym,
            ipsae_d0dom_asym, ipsae_d0res_asym
        """

        # Initialize dictionaries
        iptm_d0chn_byres = self.nested_chain_map(
            lambda: np.zeros(self.nres), self.chain_ids)
        ipsae_d0chn_byres = self.nested_chain_map(
            lambda: np.zeros(self.nres), self.chain_ids)
        ipsae_d0dom_byres = self.nested_chain_map(
            lambda: np.zeros(self.nres), self.chain_ids)
        ipsae_d0res_byres = self.nested_chain_map(
            lambda: np.zeros(self.nres), self.chain_ids)

        iptm_d0chn_asym = self.nested_chain_map(lambda: 0.0, self.chain_ids)
        ipsae_d0chn_asym = self.nested_chain_map(lambda: 0.0, self.chain_ids)
        ipsae_d0dom_asym = self.nested_chain_map(lambda: 0.0, self.chain_ids)
        ipsae_d0res_asym = self.nested_chain_map(lambda: 0.0, self.chain_ids)

        # Track unique residues
        unique_residues_chain1 = self.nested_chain_map(set, self.chain_ids)
        unique_residues_chain2 = self.nested_chain_map(set, self.chain_ids)
        dist_unique_residues_chain1 = self.nested_chain_map(set, self.chain_ids)
        dist_unique_residues_chain2 = self.nested_chain_map(set, self.chain_ids)

        # Compute per-residue iPTM / IPSAE
        for chain1 in self.chain_ids:
            for chain2 in self.chain_ids:
                if chain1 == chain2:
                    continue

                n0chn = np.sum(self.chain_res_map == chain1) + \
                    np.sum(self.chain_res_map == chain2)
                d0chn = self.calc_d0(n0chn)

                ptm_matrix_d0chn = self.ptm_func(self.pae_matrix, d0chn)

                valid_pairs_iptm = (self.chain_res_map == chain2)
                valid_pairs_matrix = (self.chain_res_map == chain2) & \
                    (self.pae_matrix < self.pae_cutoff)

                for i in range(self.nres):
                    if self.chain_res_map[i] != chain1:
                        continue

                    # IPSAE / iPTM per residue
                    valid_pairs_ipsae = valid_pairs_matrix[i]
                    iptm_d0chn_byres[chain1][chain2][i] = (
                        ptm_matrix_d0chn[
                            i, valid_pairs_iptm
                        ].mean() if valid_pairs_iptm.any() else 0.0
                    )
                    ipsae_d0chn_byres[chain1][chain2][i] = (
                        ptm_matrix_d0chn[
                            i, valid_pairs_ipsae
                        ].mean() if valid_pairs_ipsae.any() else 0.0
                    )

                    # Track unique residues contributing
                    if valid_pairs_ipsae.any():
                        unique_residues_chain1[chain1][chain2].add(i)
                        for j in np.where(valid_pairs_ipsae)[0]:
                            unique_residues_chain2[chain1][chain2].add(j)

                    valid_pairs = (self.chain_res_map == chain2) & \
                        (self.pae_matrix[i] < self.pae_cutoff) & \
                        (distances[i] < self.distance_cutoff)
                    if valid_pairs.any():
                        dist_unique_residues_chain1[chain1][chain2].add(i)
                        for j in np.where(valid_pairs)[0]:
                            dist_unique_residues_chain2[chain1][chain2].add(j)

        for chain1 in self.chain_ids:
            for chain2 in self.chain_ids:
                if chain1 == chain2:
                    continue

                # Domain-level d0
                residues_1 = len(unique_residues_chain1[chain1][chain2])
                residues_2 = len(unique_residues_chain2[chain1][chain2])
                n0dom = residues_1 + residues_2
                d0dom = self.calc_d0(n0dom)

                # PTM matrix for domain-level IPSAE
                ptm_matrix_d0dom = self.ptm_func(self.pae_matrix, d0dom)

                # Valid pairs matrix
                mask_chain2 = (self.chain_res_map == chain2)
                mask_pae = (self.pae_matrix < self.pae_cutoff)
                valid_pairs_matrix = mask_pae & mask_chain2[None, :]

                # Per-residue d0 for IPSAE
                n0res_byres_all = np.sum(valid_pairs_matrix, axis=1)
                d0res_byres_all = self.calc_d0(n0res_byres_all)

                for i in range(self.nres):
                    if self.chain_res_map[i] != chain1:
                        continue

                    valid_pairs = valid_pairs_matrix[i]

                    ipsae_d0dom_byres[chain1][chain2][i] = ptm_matrix_d0dom[
                        i, valid_pairs
                    ].mean() if valid_pairs.any() else 0.0

                    ptm_row_d0res = self.ptm_func(
                        self.pae_matrix[i], d0res_byres_all[i]
                    )
                    ipsae_d0res_byres[chain1][chain2][i] = ptm_row_d0res[
                        valid_pairs
                    ].mean() if valid_pairs.any() else 0.0

        # Compute per-chain-pair iPTM / IPSAE as maximum over residues
        for c1 in self.chain_ids:
            for c2 in self.chain_ids:
                if c1 == c2:
                    continue

                iptm_d0chn_asym[c1][c2] = iptm_d0chn_byres[c1][c2].max()
                ipsae_d0chn_asym[c1][c2] = ipsae_d0chn_byres[c1][c2].max()
                ipsae_d0dom_asym[c1][c2] = ipsae_d0dom_byres[c1][c2].max()
                ipsae_d0res_asym[c1][c2] = ipsae_d0res_byres[c1][c2].max()

        return {
            "iptm_d0chn_asym": iptm_d0chn_asym,
            "ipsae_d0chn_asym": ipsae_d0chn_asym,
            "ipsae_d0dom_asym": ipsae_d0dom_asym,
            "ipsae_d0res_asym": ipsae_d0res_asym,
            "unique_residues_chain1": {
                c1: {c2: len(res_set) for c2, res_set in inner.items()}
                for c1, inner in unique_residues_chain1.items()
            },
            "unique_residues_chain2": {
                c1: {c2: len(res_set) for c2, res_set in inner.items()}
                for c1, inner in unique_residues_chain2.items()
            },
            "dist_unique_residues_chain1": {
                c1: {c2: len(res_set) for c2, res_set in inner.items()}
                for c1, inner in dist_unique_residues_chain1.items()
            },
            "dist_unique_residues_chain2": {
                c1: {c2: len(res_set) for c2, res_set in inner.items()}
                for c1, inner in dist_unique_residues_chain2.items()
            }
        }

    def main(self, output_csv=None, verbose=True):
        """
        Main function to calculate all the scores output by ipSAE
        """
        distances = self.get_cb_distance()

        pdockq_scores, pdockq2_scores = self.compute_pdockq(distances)
        ligand_interface_scores = self.compute_ligand_interface_scores()
        iptm_ipsae_scores = self.compute_iptm_ipsae(distances)

        self.output_results(pdockq_scores,
                            pdockq2_scores,
                            ligand_interface_scores,
                            iptm_ipsae_scores,
                            verbose,
                            output_csv)

        return {
            "pdockq": pdockq_scores,
            "pdockq2": pdockq2_scores,
            "ligand_interface_scores": ligand_interface_scores,
            "iptm_ipsae_scores": iptm_ipsae_scores
        }

    def output_results(self,
                       pdockq_scores,
                       pdockq2_scores,
                       lis_scores,
                       ipsae_scores,
                       verbose=True,
                       output_csv=None):

        if not verbose and output_csv is None:
            return

        metric_dicts = {}
        if ipsae_scores:
            metric_dicts["ipSAE"] = ipsae_scores['ipsae_d0res_asym']
            metric_dicts["ipSAE_d0chn"] = ipsae_scores['ipsae_d0chn_asym']
            metric_dicts["ipSAE_d0dom"] = ipsae_scores['ipsae_d0dom_asym']
            metric_dicts["iPTM_d0chn"] = ipsae_scores['iptm_d0chn_asym']
        if pdockq_scores:
            metric_dicts["pDockQ"] = pdockq_scores
        if pdockq2_scores:
            metric_dicts["pDockQ2"] = pdockq2_scores
        if lis_scores:
            metric_dicts["LIS"] = lis_scores
        if ipsae_scores:
            metric_dicts['nres1'] = ipsae_scores['unique_residues_chain1']
            metric_dicts['nres2'] = ipsae_scores['unique_residues_chain2']
            metric_dicts['dist1'] = ipsae_scores['dist_unique_residues_chain1']
            metric_dicts['dist2'] = ipsae_scores['dist_unique_residues_chain2']

        rows = []
        chain_pairs = set()
        for metric, scores in metric_dicts.items():
            if scores is None:
                continue
            for ch1, inner in scores.items():
                for ch2 in inner.keys():
                    chain_pairs.add((ch1, ch2))

        # Build rows for each pair
        for ch1, ch2 in sorted(chain_pairs):
            row = {"Chn1": ch1, "Chn2": ch2}

            # Add available metrics
            for metric, scores in metric_dicts.items():
                if scores is None:
                    continue
                try:
                    row[metric] = scores[ch1][ch2]
                except KeyError:
                    row[metric] = None  # missing entry → leave blank

            rows.append(row)

        # Convert to DataFrame
        df = pd.DataFrame(rows)

        if verbose:
            logger.info(df.to_string(index=False))

        if output_csv:
            df.to_csv(output_csv, index=False)

        return df


def main():
    import argparse

    from Bio.PDB import MMCIFIO, PDBParser

    parser = argparse.ArgumentParser(
        prog="ipsae",
        description="Compute IPSAE metrics from structure and PAE file"
    )

    parser.add_argument("input_model", help="Path to input structure (PDB/mmCIF)")
    parser.add_argument("pae_file", help="Path to PAE file (npz/npy/json)")
    parser.add_argument("--pae_cutoff", type=float, default=10.0,
                        help="PAE cutoff used in some scores (default: 10.0)")
    parser.add_argument("--pae_format",
                        choices=[
                            "alphafold2", "alphafold3", "boltz",
                            "chai", "colabfold", "protenix"
                        ],
                        default="alphafold3",
                        help="Format of the PAE file (default: alphafold3)")
    parser.add_argument("--distance_cutoff", type=float, default=10.0,
                        help="Cβ distance cutoff in Ångström (default: 10.0)")
    parser.add_argument("--quiet", "-q", action="store_true",
                        help="Enable verbose logging/output")
    parser.add_argument("--output", "-o", type=Path,
                        help="Optional output CSV path for results")

    args = parser.parse_args()

    tmp_cif = None
    pdb_path = Path(args.input_model)
    if pdb_path.suffix == ".pdb":
        pdb_parser = PDBParser(QUIET=True)
        structure = pdb_parser.get_structure(pdb_path.stem, str(pdb_path))
        io = MMCIFIO()
        io.set_structure(structure)
        tmp_cif = pdb_path.with_suffix(".cif")
        io.save(str(tmp_cif))
        args.input_model = tmp_cif

    ipsae_calculator = Ipsae(
        args.input_model,
        args.pae_file,
        args.pae_cutoff,
        args.pae_format,
        args.distance_cutoff
    )

    if args.quiet:
        verbose = False
    else:
        verbose = True

    ipsae_calculator.main(verbose=verbose, output_csv=args.output)

    if tmp_cif is not None:
        try:
            Path(tmp_cif).unlink()
        except FileNotFoundError:
            pass


if __name__ == "__main__":
    main()



import argparse
import logging
import sys
from pathlib import Path
from textwrap import dedent

logger = logging.getLogger("logger")


def validate_json_file(value):
    """
    Validate that the input is a JSON file with a .json suffix.
    """
    if not value.endswith(".json"):
        raise argparse.ArgumentTypeError(
            f"Input file must have a .json suffix: {value}"
        )
    if not Path(value).exists():
        raise argparse.ArgumentTypeError(f"Input file does not exist: {value}")
    return value


def main_argpase_util(parser):
    parser.add_argument(
        "input_json",
        type=validate_json_file,
        help="Path to the input JSON in AlphaFold3 format",
    )
    parser.add_argument("output_dir", help="Path to the output directory")
    parser.add_argument(
        "--override",
        help="[optional] Override the existing output directory, if it exists",
        action="store_true",
    )
    parser.add_argument(
        "--output_json",
        help=dedent("[optional] Specify the path of the output ABCFold json file, this \
        can be used to run subsequent runs of ABCFold with the same input features \
        (e.g. MSA)"),
    )

    return parser


def mmseqs2_argparse_util(parser):
    parser.add_argument(
        "--mmseqs2",
        action="store_true",
        help=dedent("[optional] Use MMseqs2 for MSA generation and template \
        searching (if used with --templates flag)"),
    )
    parser.add_argument(
        "--mmseqs_database",
        help=dedent("[optional] The database directory for the generation of the MSA. \
        This is only required if using a local installation of MMseqs2"),
    )
    parser.add_argument(
        "--templates", action="store_true", help="[optional] Enable template search"
    )
    parser.add_argument(
        "--num_templates",
        type=int,
        default=20,
        help="[optional] The number of templates to use (default: 20)",
    )

    return parser


def custom_template_argpase_util(parser):
    parser.add_argument(
        "--target_id",
        nargs="+",
        help=dedent("[conditionally required] The ID of the sequence that the \
        custom template relates to. This is only required if modelling a complex. \
        If providing a list of custom templates, the target_id must be a list of \
        the same length as the custom template list"),
    )
    parser.add_argument(
        "--custom_template",
        nargs="+",
        help=dedent("[optional] Path to a custom template file in mmCif format or a \
        list of paths to custom template files in mmCif format. If providing a list of \
        custom templates, you must also provide a list of custom template chains."),
    )
    parser.add_argument(
        "--custom_template_chain",
        nargs="+",
        help=dedent("[conditionally required] The chain ID of the chain to use in your \
        custom template. This is only required if using a multi-chain template. If \
        providing a list of custom templates, you must also provide a list of custom \
        template chains of the same length as the custom template list"),
    )

    return parser


def prediction_argparse_util(parser):
    parser.add_argument(
        "--number_of_models",
        type=int,
        default=5,
        help=dedent("[optional] The number of models to generate with each method \
        (default: 5)"),
    )
    parser.add_argument(
        "--num_recycles",
        type=int,
        default=10,
        help="[optional] Number of recycles to use during inference (default: 10)",
    )
    return parser


def boltz_argparse_util(parser):
    parser.add_argument(
        "-b",
        "--boltz",
        action="store_true",
        help="Run Boltz",
    )
    if "--save_input" not in parser._option_string_actions:
        parser.add_argument(
            "--save_input",
            action="store_true",
            help="Save the input json file",
            default=False,
        )

    return parser


def chai_argparse_util(parser):
    parser.add_argument(
        "-c",
        "--chai1",
        action="store_true",
        help="Run Chai-1",
    )
    return parser


def protenix_argparse_util(parser):
    parser.add_argument(
        "-p",
        "--protenix",
        action="store_true",
        help="Run Protenix",
    )
    return parser


def alphafold_argparse_util(parser):
    parser.add_argument(
        "--database",
        help=dedent("[optional] The database directory for the generation of the MSA. \
        This is only required if using the built in AlphaFold3 MSA generation"),
        dest="database_dir",
        default=None,
    )

    parser.add_argument(
        "--model_params",
        help="[required] The directory containing the AlphaFold3 model parameters",
        default=None,
    )

    parser.add_argument(
        "--af3_sif_path",
        help=dedent("[conditionally required] The path to the sif image of AlphaFold3 \
        if using Singularity"),
        default=None,
    )

    parser.add_argument(
        "-a",
        "--alphafold3",
        action="store_true",
        help="Run Alphafold3",
    )

    parser.add_argument(
        "--use_af3_template_search",
        action="store_true",
        help=dedent("If providing your own custom MSA or if you've run `--mmseqs2`, \
        allow Alphafold3 to search for templates"),
    )

    parser.add_argument(
        "--save_distogram",
        action="store_true",
        help="[optional] store AlphaFold3 distograms",
    )

    return parser


def visuals_argparse_util(parser):
    parser.add_argument(
        "--no_visuals",
        action="store_true",
        help=dedent("[optional] Do not generate the output pages, best for running on \
        a cluster without a display"),
    )

    parser.add_argument(
        "--no_server",
        action="store_true",
        help=dedent("[optional] Do not start a local server to view the results, the \
        output page is still generated and is accessible in the output directory"),
    )
    return parser


def raise_argument_errors(args):
    if not args.alphafold3 and not args.boltz and not args.chai1 and not args.protenix:
        logger.info(
            dedent("Neither AlphaFold3, Boltz, Chai-1, or Protenix selected. \
            Running AlphaFold3 by default")
        )
        args.alphafold3 = True

    if (
        args.alphafold3
        and (not args.model_params or not Path(args.model_params).exists())
        and not args.mmseqs2
    ):
        logger.error(f"Model parameters directory not found: {args.model_params}")
        sys.exit(1)

    if args.templates and not args.mmseqs2 and not args.alphafold3:
        logger.error("Cannot use --templates flag without using MMseqs2 or Alphafold3")
        sys.exit(1)

    if (
        args.templates
        and args.alphafold3
        and not args.mmseqs2
        and not args.use_af3_template_search
    ):
        # Ensure templates are used with Alphafold3 if --templates is set
        args.use_af3_template_search = True

    if args.custom_template_chain and not args.custom_template:
        logger.error("Custom template chain provided without a custom template")
        sys.exit(1)

    if args.use_af3_template_search and not args.alphafold3:
        logger.error(
            "Cannot use the Alphafold3 template search without running Alphafold3"
        )
        sys.exit(1)

    if args.num_templates < 1:
        logger.error("Number of templates must be greater than 0")
        sys.exit(1)

    if args.num_recycles < 1:
        logger.error("Number of recycles must be greater than 0")
        sys.exit(1)

    if args.number_of_models < 1:
        logger.error("Number of models must be greater than 0")
        sys.exit(1)

    return args



import logging
import shutil
from pathlib import Path
from typing import Union

from abcfold.chai1.af3_to_chai import ChaiFasta
from abcfold.output.file_handlers import (CifFile, ConfidenceJsonFile,
                                          FileTypes, NpyFile, NpzFile)
from abcfold.output.utils import Af3Pae

logger = logging.getLogger("logger")


class ChaiOutput:
    def __init__(
        self,
        chai_output_dirs: list[Union[str, Path]],
        input_params: dict,
        name: str,
        save_input: bool = False,
    ):
        """
        Object to process the output of an Chai-1 run

        Args:
            chai_output_dirs (list[Union[str, Path]]): Path to the Chai-1
            output directory
            input_params (dict): Dictionary containing the input parameters used for the
            Chai-1 run
            name (str): Name given to the Chai-1 run
            save_input (bool): If True, Chai-1 was run with the save_input flag

        Attributes:
            input_params (dict): Dictionary containing the input parameters used for the
            Chai-1 run
            output_dirs (Path):  List of paths to the Chai-1 output directory(s)
            name (str): Name given to the Chai-1 run
            output (dict): Dictionary containing the processed output the contents
            of the Chai-1 output directory(s). The dictionary is structured as follows:

            {
                "seed-1": {
                    1: {
                        "pae": NpzFile,
                        "cif": CifFile,
                        "scores": NpyFile
                    },
                    2: {
                        "pae": NpzFile,
                        "cif": CifFile,
                        "scores": NpyFile
                    },
                etc...
            }
            pae_files (list): Ordered list of NpzFile objects containing the PAE data
            cif_files (list): Ordered list of CifFile objects containing the CIF data
            scores_files (list):  Ordered list of NpyFile objects containing the scores
            data

        """
        self.input_params = input_params
        self.output_dirs = [Path(x) for x in chai_output_dirs]
        self.name = name
        self.save_input = save_input

        parent_dir = self.output_dirs[0].parent
        new_parent = parent_dir / f"chai1_{self.name}"
        new_parent.mkdir(parents=True, exist_ok=True)

        if self.save_input:
            chai_fasta = parent_dir / "chai1.fasta"
            if chai_fasta.exists():
                chai_fasta.rename(new_parent / "chai1.fasta")
            chai_msas = list(parent_dir.glob("*.aligned.pqt"))
            if chai_msas:
                for chai_msa in chai_msas:
                    if chai_msa.exists():
                        chai_msa.rename(new_parent / chai_msa.name)

        new_output_dirs = []
        for output_dir in self.output_dirs:
            if output_dir.name.startswith("chai_output_"):
                new_path = new_parent / output_dir.name
                output_dir.rename(new_path)
                new_output_dirs.append(new_path)
            else:
                new_output_dirs.append(output_dir)
        self.output_dirs = new_output_dirs

        self.input_fasta = self.get_input_fasta()

        self.output = self.process_chai_output()
        self.seeds = list(self.output.keys())

        self.pae_files = {
            seed: [value["pae"] for value in self.output[seed].values()
                   if "pae" in value] for seed in self.seeds
        }
        self.cif_files = {
            seed: [value["cif"] for value in self.output[seed].values()]
            for seed in self.seeds
        }
        self.scores_files = {
            seed: [value["scores"] for value in self.output[seed].values()]
            for seed in self.seeds
        }
        self.pae_to_af3()
        self.af3_pae_files = {
            seed: [value["af3_pae"] for value in self.output[seed].values()]
            for seed in self.seeds
        }

    def process_chai_output(self):
        file_groups = {}

        for pathway in self.output_dirs:
            seed = pathway.name.split("_")[-1]
            if seed not in file_groups:
                file_groups[seed] = {}

            for output in pathway.rglob("*"):
                number = output.stem.split("model_idx_")[-1]
                if number.isdigit():
                    number = int(number)

                file_type = output.suffix[1:]

                if file_type == FileTypes.NPZ.value:
                    file_ = NpzFile(str(output))

                elif file_type == FileTypes.CIF.value:
                    file_ = CifFile(str(output), self.input_params)
                    file_ = self.update_chain_labels(file_)

                elif file_type == FileTypes.NPY.value:
                    file_ = NpyFile(str(output))
                else:
                    continue

                if isinstance(number, str):
                    number = -1

                if number not in file_groups[seed]:
                    file_groups[seed][number] = [file_]
                else:
                    file_groups[seed][number].append(file_)

        seed_dict = {}
        for seed, models in file_groups.items():
            model_number_file_type_file = {}
            pae_file = None
            if -1 in models:
                for file_ in models[-1]:
                    if file_.pathway.stem.startswith("pae_scores"):
                        pae_file = file_
                        break

            for model_number, files in models.items():
                if model_number == -1:
                    continue
                intermediate_dict = {}
                for file_ in sorted(files, key=lambda x: x.suffix):
                    if file_.pathway.stem.startswith("scores.model"):
                        intermediate_dict["scores"] = file_
                    elif file_.pathway.stem.startswith("pred.model"):
                        file_.name = f"Chai-1_{seed}_{model_number}"
                        # Chai cif not recognised by pae-viewer, so we load and save
                        file_.to_file(file_.pathway)
                        intermediate_dict["cif"] = file_
                if model_number != -1 and pae_file is not None:
                    new_pae_path = (
                        file_.pathway.parent / f"pae_scores_model_{model_number}.npy"
                    )
                    shutil.copy(pae_file.pathway, new_pae_path)
                    intermediate_dict["pae"] = NpyFile(str(new_pae_path))

                model_number_file_type_file[model_number] = intermediate_dict

            model_number_file_type_file = {
                model_number: model_number_file_type_file[model_number]
                for model_number in sorted(model_number_file_type_file)
            }
            seed_dict[seed] = model_number_file_type_file

        return seed_dict

    def pae_to_af3(self):
        """
        Convert the Chai-1 PAE data to the format expected by AlphaFold3

        Returns:
            None
        """
        new_pae_files = {}
        for seed in self.seeds:
            for i, (pae_file, cif_file) in enumerate(
                zip(self.pae_files[seed], self.cif_files[seed])
            ):
                pae = Af3Pae.from_chai1(
                    pae_file.data[i],
                    cif_file,
                )

                out_name = pae_file.pathway

                pae.to_file(out_name)

                if seed not in new_pae_files:
                    new_pae_files[seed] = []
                new_pae_files[seed].append(ConfidenceJsonFile(out_name))

        self.output = {
            seed: {
                i: {
                    "cif": cif_file,
                    "af3_pae": new_pae_files[seed][i],
                    "scores": self.output[seed][i]["scores"],
                }
                for i, cif_file in enumerate(self.cif_files[seed])
            }
            for seed in self.seeds
        }

    def get_input_fasta(self) -> ChaiFasta:
        """
        Function to get the input fasta file used for the Chai-1 run

        Returns:
            ChaiFasta: ChaiFasta object containing the input fasta file

        """

        ch = ChaiFasta(self.output_dirs[0], create_files=False)
        ch.json_to_fasta(self.input_params)

        return ch

    def update_chain_labels(self, cif_file: CifFile) -> CifFile:
        """
        Function to update the chain labels in the CIF file

        Args:
            cif_file (CifFile): CifFile object to update the chain labels for

        """

        cif_file.relabel_chains(self.input_fasta.chain_ids)
        cif_file.to_file(cif_file.pathway)
        return CifFile(cif_file.pathway, self.input_params)
